import numpy as np
from calng.stacking_utils import StackingFriend
from calngUtils import device as device_utils
from .utils import kwhash
from karabo.bound import (
    IMAGEDATA_ELEMENT,
    KARABO_CLASSINFO,
    NDARRAY_ELEMENT,
    INT32_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    Hash,
    ImageData,
    Schema,
)
from TrainMatcher import TrainMatcher

from . import scenes
from ._version import version as deviceVersion


@device_utils.with_unsafe_get
@KARABO_CLASSINFO("AzimuthalCombiner", deviceVersion)
class AzimuthalCombiner(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            .setNewDefaultValue(["overview", "trainMatcherScene"])
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("sortSources")
            .setNowReadOnly()
            .setNewDefaultValue(False)
            .commit(),
        )
        StackingFriend.add_schema(expected)
        (
            OVERWRITE_ELEMENT(expected)
            .key("merge")
            .setNewDefaultValue(
                [
                    kwhash(
                        select=True,
                        sourcePattern=r".*/DET/.*:(xtdf|daqOutput)",
                        keyPattern=key,
                        replacement="STACKED",
                        groupType="sources",
                        mergeMethod="stack",
                        axis=0,
                        missingValue="nan",
                    )
                    for key in ("azimuthal.integrated1d", "azimuthal.integrated1dWeight", "azimuthal.bins")
                ]
            )
            .commit(),
        )
        output_schema = Schema()
        preview_streak_schema = Schema()
        preview_hist_schema = Schema()
        (
            NDARRAY_ELEMENT(output_schema)
            .key("integrated1d")
            .commit(),

            NDARRAY_ELEMENT(output_schema)
            .key("bins")
            .commit(),

            IMAGEDATA_ELEMENT(preview_streak_schema)
            .key("image")
            .commit(),

            NDARRAY_ELEMENT(preview_hist_schema)
            .key("x")
            .commit(),

            NDARRAY_ELEMENT(preview_hist_schema)
            .key("y")
            .commit(),
        )
        (
            OUTPUT_CHANNEL(expected)
            .key("output")
            .dataSchema(output_schema)
            .commit(),

            # emulating the schema of devices using preview friend
            NODE_ELEMENT(expected)
            .key("preview")
            .commit(),

            NODE_ELEMENT(expected)
            .key("preview.streak")
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("preview.streak.output")
            .dataSchema(preview_streak_schema)
            .commit(),

            NODE_ELEMENT(expected)
            .key("preview.hist")
            .commit(),

            INT32_ELEMENT(expected)
            .key("preview.hist.index")
            .displayedName("Preview index")
            .description(
                "In case of multi-frame data, which frame should we preview? Either "
                "select frame by using value >= 0 or reduction with special values: "
                "-1: max, -2: mean, -3: sum, -4: stdev. If you specify a frame out "
                "of range, I will silently clamp index to some legal value."
            )
            .assignmentOptional()
            .defaultValue(0)
            .minInc(-4)
            .reconfigurable()
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("preview.hist.output")
            .dataSchema(preview_hist_schema)
            .commit(),
        )

    def initialization(self):
        super().initialization()
        self._stacking_friend = StackingFriend(
            self, self.get("sources"), self.get("merge")
        )
        self.start()

    def on_matched_data(self, tid, sources):
        # snatch up bins from arbitrary source
        # TODO: maybe check that bins agree
        for source, (data, _) in sources.items():
            if data.has("azimuthal.bins"):
                bins = data["azimuthal.bins"]
                break
        else:
            self.log.WARN("No source provided azimuthal integration bins")
            return
        bin_midpoints = (bins[1:] + bins[:-1]) / 2
        # collect actual results
        self._stacking_friend.process(sources)
        data_hash, _ = sources["STACKED"]

        # combining data across modules
        combined = np.nansum(data_hash["azimuthal.integrated1d"], axis=0) / np.nansum(
            data_hash["azimuthal.integrated1dWeight"], axis=0
        )

        # regular output: "full" combined data + actual bin edges
        self.writeChannel("output", Hash("bins", bins, "integrated1d", combined))

        # streak preview: same data, just as image
        self.writeChannel(
            "preview.streak.output",
            Hash("image", ImageData(np.nan_to_num(combined)))
        )

        # histogram: select some frame
        preview_index = self.unsafe_get("preview.hist.index")
        if preview_index == -1:
            preview_data = np.nanmax(combined, axis=0)
        elif preview_index == -2:
            preview_data = np.nanmean(combined, axis=0)
        elif preview_index == -3:
            preview_data = np.nansum(combined, axis=0)
        elif preview_index == -4:
            preview_data = np.nanstd(combined, axis=0)
        else:
            preview_index = max(0, min(preview_index, combined.shape[0]-1))
            preview_data = combined[preview_index]
        np.nan_to_num(preview_data, copy=False)
        self.writeChannel(
            "preview.hist.output",
            Hash(
                "x", bin_midpoints,
                "y", preview_data,
            )
        )

        self.rate_out.update()
        self.info["sent"] += 1
        self.info["trainId"] = tid

    def requestScene(self, params):
        scene_name = params.get("name", default="")
        if scene_name == "trainMatcherScene":
            params["name"] = "scene"
            return super().requestScene(params)
        if scene_name != "overview":
            self.log.INFO(
                f"Unexpected scene name {scene_name} requested, "
                "will just return overview"
            )
        payload = Hash("name", scene_name, "success", True)
        payload["data"] = scenes.azimuthal_combiner_overview(
            device_id=self.getInstanceId(),
            schema=self.getFullSchema(),
        )
        self.reply(
            Hash(
                "type",
                "deviceScene",
                "origin",
                self.getInstanceId(),
                "payload",
                payload,
            )
        )
