from karabo.bound import (
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
)
import numpy as np
from .base_kernel import BaseArbiterKernel


class BooleanCombination(BaseArbiterKernel):
    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            OVERWRITE_ELEMENT(schema)
            .key(prefix)
            .setNewDescription(
                "This kernel allows combining individual masks from many sources. It "
                "is intended for cases where, for example, each correction device via "
                "some addon provides its own frame mask. This arbiter kernel would "
                "then be able to AND (keep only frames making it through all masks) or "
                "OR these individual masks."
            )
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.key")
            .assignmentOptional()
            .defaultValue("data.dataFramePattern")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.operator")
            .assignmentOptional()
            .defaultValue("all")
            .options("all,any")
            .reconfigurable()
            .commit(),
        )

    def consider(self, train_id, sources, num_frames, mask, out_hash):
        # pretty sure this is special case of reduce and threshold in some algebra
        res = np.zeros(num_frames, dtype=bool)
        mask = mask.astype(bool, copy=False)
        sources_with_key = [
            data[self._key][mask]
            for (data, _) in sources.values()
            if data.has(self._key)
        ]
        if not sources_with_key:
            raise KeyError(f"No sources had '{self._key}'")
        res[mask] = self._operator(sources_with_key, axis=0)
        return res

    def reconfigure(self, config):
        if config.has("key"):
            self._key = config.get("key")
        if config.has("operator"):
            self._operator = getattr(np, config.get("operator"))
