import enum
import numpy as np
from karabo.bound import STRING_ELEMENT

from .base_kernel import BaseArbiterKernel


class FramefilterSpecType(enum.Enum):
    RANGE = "range"
    MASK = "mask"
    INDICES = "indices"


class ManualFilter(BaseArbiterKernel):
    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            STRING_ELEMENT(schema)
            .key(f"{prefix}.resize")
            .description(
                "If number of frames doesn't match the length of the specified "
                "filter, what to do? If 'Error', the kernel will just fail (resulting "
                "in all true for the arbiter). If True or False, will expand manual "
                "filter with the corresponding value (shrinking if manual filter is "
                "too big)."
            )
            .options("True,False,Error")
            .assignmentOptional()
            .defaultValue("False")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.spec")
            .description(
                "String specifying the manual mask to apply. See 'type' for details."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.type")
            .description(
                "How to specify the manual filter. 'range' means parse and pass the "
                "spec to np.arange (so should be between 1 and 3 comma-separated "
                "integers specifying [start], stop, [step]). 'mask' means "
                "comma-separated list of 0 and 1 (zero means don't include, 1 means "
                "include). 'indices' means comma-separated list of indices (these will "
                "be included, anything else will not be)."
            )
            .options(",".join(ffst.value for ffst in FramefilterSpecType))
            .assignmentOptional()
            .defaultValue("mask")
            .reconfigurable()
            .commit(),
        )

    def reconfigure(self, config):
        # TODO: handle rollback if illegal config
        if config.has("type"):
            self._filter_type = FramefilterSpecType(config["type"])
        if config.has("spec"):
            self._filter_spec = config["spec"]
        if config.has("spec") or config.has("type"):
            self._filter = _parse_frame_filter(self._filter_type, self._filter_spec)

        if config.has("resize"):
            if config["resize"] == "Error":

                def aux(num_frames):
                    if num_frames != self._filter.size:
                        raise ValueError("Number of frames doesn't match my filter")
                    return self._filter

            else:
                value = config["resize"] == "True"  # note: can't just use bool(...)

                def aux(num_frames):
                    if num_frames < self._filter.size:
                        return self._filter[:num_frames]
                    return np.pad(
                        self._filter,
                        (0, num_frames - self._filter.size),
                        constant_values=value,
                    )

            self._resize = aux

    def consider(self, train_id, sources, num_frames, mask, out_hash):
        return mask & self._resize(num_frames)


def _parse_frame_filter(filter_type, filter_spec):
    if filter_type in (FramefilterSpecType.RANGE, FramefilterSpecType.INDICES):
        if filter_type is FramefilterSpecType.RANGE:
            # allow exceptions to bubble up
            numbers = tuple(int(part) for part in filter_spec.split(","))
            indices = np.arange(*numbers, dtype=np.uint16)
        else:
            indices = np.fromstring(filter_spec, dtype=np.uint16, sep=",")
        mask = np.zeros(np.max(indices)+1, dtype=bool)
        mask[indices] = True
    elif filter_type is FramefilterSpecType.MASK:
        mask = np.fromstring(filter_spec, dtype=bool, sep=",")
    else:
        raise TypeError(f"Unknown frame filter type {filter_type}")
    return mask
