import copy
import itertools
import scipy.sparse

import numpy as np

from calngUtils.misc import ceil_div

from .base_kernel_runner import get_kernel_template
from .utils import np_dtype_to_c_type
import calng.kernels.azimuthal_cython as az


def slice_geom(geom, modules):
    if isinstance(modules, int):
        modules = [modules]
    res = copy.copy(geom)
    res.modules = [res.modules[i] for i in modules]
    res.expected_data_shape = (len(modules),) + res.expected_data_shape[1:]
    res.n_modules = len(res.modules)
    return res


def get_geom_tile_vectors(geom, det_dist):
    num_tiles = geom.n_tiles_per_module * geom.n_modules
    tile_corners = np.empty((num_tiles, 3), dtype=np.float64)
    tile_ss_vecs = np.empty((num_tiles, 3), dtype=np.float64)
    tile_fs_vecs = np.empty((num_tiles, 3), dtype=np.float64)
    for i, tile in enumerate(itertools.chain.from_iterable(geom.modules)):
        assert tile.ss_pixels == geom.frag_ss_pixels
        assert tile.fs_pixels == geom.frag_fs_pixels
        tile_corners[i] = tile.corner_pos
        tile_ss_vecs[i] = tile.ss_vec
        tile_fs_vecs[i] = tile.fs_vec
    tile_corners[:, -1] += det_dist
    return tile_corners, tile_ss_vecs, tile_fs_vecs


def get_geom_max_theta(geom, det_dist):
    def tile_max_theta(tile):
        corner = tile.corner_pos + [0, 0, det_dist]
        return max(
            np.arctan2(np.sqrt(arr[0] ** 2 + arr[1] ** 2), arr[2])
            for arr in [
                corner,
                corner + tile.ss_vec * tile.ss_pixels,
                corner + tile.fs_vec * tile.fs_pixels,
                corner + tile.ss_vec * tile.ss_pixels + tile.fs_vec * tile.fs_pixels,
            ]
        )

    return max(
        tile_max_theta(tile) for tile in itertools.chain.from_iterable(geom.modules)
    )


def coo_from_geom(geom, bins, det_dist, n_split):
    tile_corners, tile_ss_vecs, tile_fs_vecs = get_geom_tile_vectors(geom, det_dist)
    data, rows, cols = az.generate_coo(
        bins,
        tile_corners,
        tile_ss_vecs,
        tile_fs_vecs,
        geom.frag_ss_pixels,
        geom.frag_fs_pixels,
        n_split,
    )
    return scipy.sparse.coo_matrix(
        (
            data,
            (rows, cols)
        ),
        dtype=np.float32,
        shape=(bins.size - 1, np.prod(geom.expected_data_shape)),
    )


def coo_to_sorted_csr(coo):
    order = np.lexsort((coo.row, coo.col))
    return scipy.sparse.coo_matrix(
        (
            coo.data[order],
            (
                coo.row[order],
                coo.col[order],
            ),
        ),
        shape=coo.shape,
    ).tocsr()


class BaseAzimuthalIntegrator:
    def __init__(self, geom, bins, det_dist, n_split):
        self.geom = geom
        if isinstance(bins, int):
            self.bins = np.linspace(0, get_geom_max_theta(geom, det_dist), bins)
        else:
            self.bins = np.asarray(bins, dtype=np.float32)
        self.det_dist = det_dist
        self.n_split = n_split
        self.n_pixels = np.prod(self.geom.expected_data_shape)

    def _check_shape_make_buffer(self, data, out, out_weight, multi_module):
        if data.ndim == 1:
            if data.size != self.n_pixels:
                raise ValueError(
                    "Flat input data should have size equal to number of pixels"
                )
            n_frames = 1
            was_single_frame = True
        elif data.ndim == 2:
            if data.shape[1] != self.n_pixels:
                raise ValueError("2d data input should have shape (frames, pixels)")
            n_frames = data.shape[0]
            was_single_frame = False
        elif data.ndim == len(self.geom.expected_data_shape):
            n_frames = 1
            was_single_frame = True
            data = data.ravel()
        elif data.ndim == len(self.geom.expected_data_shape) + 1:
            if data.shape[1:] != self.geom.expected_data_shape:
                raise ValueError("nd data should have frames + geometry expected shape")
            n_frames = data.shape[0]
            data = data.reshape(n_frames, -1)
            was_single_frame = False

        output_shape = (n_frames, self.bins.size - 1)
        if out is None:
            out = self._xp.empty(output_shape, dtype=data.dtype)
        else:
            assert out.dtype == data.dtype
            assert out.shape == output_shape
        if multi_module and out_weight is None:
            out_weight = self._xp.empty_like(out)
        if was_single_frame:
            data = data[None]  # add pretend singleton axis so kernel doesn't get mad

        return data, out, out_weight, was_single_frame


class CpuAzimuthalIntegrator(BaseAzimuthalIntegrator):
    _xp = np

    def __init__(self, geom, bins, det_dist, n_split):
        super().__init__(geom, bins, det_dist, n_split)
        self.csr = coo_to_sorted_csr(
            coo_from_geom(self.geom, self.bins, self.det_dist, self.n_split)
        )

    def integrate_1d(self, data, out=None):
        data, out, _, was_single_frame = self._check_shape_make_buffer(
            data, out, None, False
        )

        az.integrate_1d_csr(
            self.csr.indices,
            self.csr.indptr,
            self.csr.data,
            data,
            out,
        )

        # just being polite: if no frame axis on input, don't give one on output
        if was_single_frame:
            out = out[0]

        return out

    def integrate_1d_multimodule(self, data, out=None, out_weight=None):
        data, out, out_weight, was_single_frame = self._check_shape_make_buffer(
            data, out, out_weight, True
        )

        az.integrate_1d_csr_multimodule(
            self.csr.indices,
            self.csr.indptr,
            self.csr.data,
            data,
            out,
            out_weight,
        )

        # just being polite: if no frame axis on input, don't give one on output
        if was_single_frame:
            out = out[0]
            out_weight = out_weight[0]

        return out, out_weight


class GpuAzimuthalIntegrator(BaseAzimuthalIntegrator):
    def __init__(self, geom, bins, det_dist, n_split):
        import cupy
        import cupyx.scipy.sparse

        self._xp = cupy
        self._sparse = cupyx.scipy.sparse
        super().__init__(geom, bins, det_dist, n_split)

        self._kernels = {}
        name_expressions = [
            "integrate_1d_csr<float,float>",
            "integrate_1d_csr_multimodule<float,float>"
        ]
        self._module = self._xp.RawModule(
            code=get_kernel_template("azimuthal.cu").render(),
            # backend="nvrtc",
            options=("-std=c++11",),
            name_expressions=name_expressions,
        )
        self._regular_kernel = self._module.get_function(name_expressions[0])
        self._multimod_kernel = self._module.get_function(name_expressions[1])

        csr_cpu = coo_to_sorted_csr(
            coo_from_geom(self.geom, self.bins, self.det_dist, self.n_split)
        )
        self.csr = self._sparse.csr_matrix(csr_cpu.astype(np.float32))

    def integrate_1d(self, data, out=None, block=(128, 8)):
        data, out, _, was_single_frame = self._check_shape_make_buffer(
            data, out, None, False
        )

        n_frames = data.shape[0]
        grid = (
            ceil_div(n_frames, block[0]),  # x is frame
            ceil_div(self.csr.shape[0], block[1]),  # y is ring
        )
        self._regular_kernel(
            grid,
            block,
            (
                *self.csr.shape,
                self.csr.data,
                self.csr.indices,
                self.csr.indptr,
                n_frames,
                data,
                out,
            ),
        )

        if was_single_frame:
            out = out[0]

        return out

    def integrate_1d_multimodule(self, data, out=None, out_weight=None, block=(128, 8)):
        data, out, out_weight, was_single_frame = self._check_shape_make_buffer(
            data, out, out_weight, True
        )
        n_frames = data.shape[0]
        grid = (
            ceil_div(n_frames, block[0]),  # x is frame
            ceil_div(self.csr.shape[0], block[1]),  # y is ring
        )
        self._multimod_kernel(
            grid,
            block,
            (
                *self.csr.shape,
                self.csr.data,
                self.csr.indices,
                self.csr.indptr,
                n_frames,
                data,
                out,
                out_weight,
            ),
        )

        if was_single_frame:
            out = out[0]
            out_weight = out_weight[0]

        return out, out_weight
