import enum
import functools
import itertools
import pathlib

from karabo.bound import (
    BOOL_ELEMENT,
    NODE_ELEMENT,
    STRING_ELEMENT,
    Hash,
)
from .utils import (
    BadPixelValues,
    maybe_get,
    subset_of_hash,
)
from calngUtils import misc
import jinja2


class BaseKernelRunner:
    runner_type = None  # subclass sets CPU / GPU
    _xp = None  # subclass sets numpy or cupy
    num_pixels_ss = None  # subclass must set
    num_pixels_fs = None  # subclass must set

    _bad_pixel_constants = None  # should be set by subclasses using bad pixels
    bad_pixel_subset = 0xFFFFFFFF  # will be set in reconfigure

    @classmethod
    def add_schema(cls, schema):
        """Using the steps in the provided mapping, will add nodes to schema

        correction_steps is assumed to be iterable of pairs where first entry in each
        pair is the name of a correction step as it will appear in device schema (second
        entry - typically an enum field - is ignored). For correction step, a node and
        some booleans are added to the schema (all tagged as managed). Subclass can
        customize / add additional keys under node later.
        """
        (
            NODE_ELEMENT(schema)
            .key("corrections")
            .commit(),
        )

        for field_name, _, used_constants in cls._correction_steps:
            node_name = f"corrections.{field_name}"
            (
                NODE_ELEMENT(schema).key(node_name).commit(),

                BOOL_ELEMENT(schema)
                .key(f"{node_name}.available")
                .displayedName("Available")
                .description(
                    "This boolean indicates whether the necessary constants have been "
                    "loaded for this correction step to be applied. Enabling the "
                    "correction will have no effect unless this is True."
                )
                .readOnly()
                # some corrections available without constants
                .initialValue(not used_constants or None in used_constants)
                .commit(),

                BOOL_ELEMENT(schema)
                .key(f"{node_name}.enable")
                .tags("managed")
                .displayedName("Enable")
                .description(
                    "Controls whether to apply this correction step for main data "
                    "output - subject to availability."
                )
                .assignmentOptional()
                .defaultValue(True)
                .reconfigurable()
                .commit(),

                BOOL_ELEMENT(schema)
                .key(f"{node_name}.preview")
                .tags("managed")
                .displayedName("Preview")
                .description(
                    "Whether to apply this correction step for corrected preview "
                    "output - subject to availability."
                )
                .assignmentOptional()
                .defaultValue(True)
                .reconfigurable()
                .commit(),
            )

    @staticmethod
    def add_bad_pixel_config(schema):
        (
            STRING_ELEMENT(schema)
            .key("corrections.badPixels.maskingValue")
            .tags("managed")
            .displayedName("Bad pixel masking value")
            .description(
                "Any pixels masked by the bad pixel mask will have their value "
                "replaced with this. Note that this parameter is to be interpreted as "
                "a numpy.float32; use 'nan' to get NaN value."
            )
            .assignmentOptional()
            .defaultValue("nan")
            .reconfigurable()
            .commit(),

            NODE_ELEMENT(schema)
            .key("corrections.badPixels.subsetToUse")
            .displayedName("Bad pixel flags to use")
            .description(
                "The booleans under this node allow for selecting a subset of bad "
                "pixel types to take into account when doing bad pixel masking. Upon "
                "updating these flags, the map used for bad pixel masking will be "
                "ANDed with this selection. Turning disabled flags back on causes "
                "reloading of cached constants."
            )
            .commit(),
        )
        for field in BadPixelValues:
            (
                BOOL_ELEMENT(schema)
                .key(f"corrections.badPixels.subsetToUse.{field.name}")
                .tags("managed")
                .assignmentOptional()
                .defaultValue(True)
                .reconfigurable()
                .commit()
            )

    def __init__(self, device):
        self._device = device
        if self.runner_type is KernelRunnerTypes.GPU:
            import os
            os.environ["CUDA_VISIBLE_DEVICES"] = str(
                self._device.unsafe_get("gpuIndex")
            )
            import cupy
            self._xp = cupy
        else:
            import numpy
            self._xp = numpy
        # for now, we depend on multiple keys from device hash, so get multiple nodes
        config = subset_of_hash(
            self._device._parameters,
            "corrections",
            "dataFormat",
            "constantParameters",
        )
        self._constant_memory_cells = config["constantParameters.memoryCells"]
        # note: does not handle one constant enabling multiple correction steps
        # (do we need to generalize "decide if this correction step is available"?)
        self._constant_to_correction_names = {}
        for (name, _, constants) in self._correction_steps:
            for constant in constants:
                self._constant_to_correction_names.setdefault(constant, set()).add(name)
        self._output_dtype = self._xp.float32
        self._setup_constant_buffers()
        self.reconfigure(config)
        self._post_init()

    def _post_init(self):
        # can be used to set GPU / CPU-specific buffers in subclasses
        # without overriding __init__
        pass

    def expected_input_shape(self, num_frames):
        """Mostly used to inform the host device about what data shape should be.
        Subclass should override as needed (ex. XTDF detectors with extra image/gain
        axis)."""
        return (num_frames, self.num_pixels_ss, self.num_pixels_fs)

    def expected_output_shape(self, num_frames):
        """Used to inform the host device about what to expect (for reshaping the shmem
        buffer). Takes into account transpose setting; override _expected_output_shape
        for subclasses which do funky shape things."""
        base_shape = self._expected_output_shape(num_frames)
        if self._output_transpose is None:
            return base_shape
        else:
            return tuple(base_shape[i] for i in self._output_transpose)

    def _expected_output_shape(self, num_frames):
        return (num_frames, self.num_pixels_ss, self.num_pixels_fs)

    def _correct(self, flags, image_data, cell_table, *rest):
        """Subclass will implement the core correcton routine with this name. Observe
        the parameter list: the "rest" part will always contain at least one output
        buffer, but may have more entries such as additional raw input buffers and
        multiple output buffers. See how it is called by correct."""

        raise NotImplementedError()

    def _preview_data_views(self, *buffers):
        """Should provide tuple of views of data for different preview layers - as a
        minimum, raw and corrected.  View is expected to have axis order frame, slow
        scan, fast scan with the latter being the final preview shape (which is X and
        which is Y may differ between detectors). The buffers passed vary from subclass
        to subclass; will be all inputs and outputs used in _correct. By default, that
        is just what gets returned."""

        return buffers

    def _load_constant(self, constant_type, constant_data):
        """Should update the appropriate internal buffers for constant type (maps to
        "calibration" type in CalCat) with the data in constant_tdata. This may include
        sanitization; data passed here has the axis order found in the stored constant
        files."""
        raise NotImplementedError()

    def _setup_constant_buffers(self):
        raise NotImplementedError()

    def _make_output_buffers(self, num_frames, flags):
        """Should, based on number of frames passed (may in the future be extended to
        include other information), create the necessary buffers to store corrected
        output(s). By default, will create a single processed data buffer of float32
        with the same shape as input image data, skipping any extra axes (like the
        image/gain axis for XTDF). Subclass should override if it wants more outputs.
        First buffer must be for corrected image data (passed to addons)"""
        return [
            self._xp.empty(
                (num_frames, self.num_pixels_ss, self.num_pixels_fs),
                dtype=self._xp.float32,
            )
        ]

    @functools.cached_property
    def _all_relevant_constants(self):
        return set(
            itertools.chain.from_iterable(
                constants for (_, _, constants) in self._correction_steps
            )
        ) - {None}

    def reconfigure(self, config):
        if config.has("constantParameters.memoryCells"):
            self._constant_memory_cells = config.get("constantParameters.memoryCells")
            self._setup_constant_buffers()
            self._device._please_send_me_cached_constants(
                self._all_relevant_constants, self.load_constant
            )

        if config.has("dataFormat.outputImageDtype"):
            self._output_dtype = self._xp.dtype(
                config.get("dataFormat.outputImageDtype")
            )

        if config.has("dataFormat.outputAxisOrder"):
            order = config["dataFormat.outputAxisOrder"]
            if order == "f-ss-fs":
                self._output_transpose = None
            else:
                self._output_transpose = misc.transpose_order(
                    ("f", "ss", "fs"), tuple(order.split("-"))
                )

        if config.has("corrections.badPixels.maskingValue"):
            self.bad_pixel_mask_value = self._xp.float32(
                config["corrections.badPixels.maskingValue"]
            )

        if config.has("corrections.badPixels.subsetToUse"):
            # note: now just always reloading from cache for convenience
            self._update_bad_pixel_subset()
            self.flush_buffers(self._bad_pixel_constants)
            self._device._please_send_me_cached_constants(
                self._bad_pixel_constants, self.load_constant
            )

        # TODO: only trigger flag update if changed (cheap though)
        self._update_correction_flags()

    def reshape(self, processed_data, out=None):
        """Move axes to desired output order"""
        if self._output_transpose is None:
            return maybe_get(processed_data, out)
        else:
            return maybe_get(processed_data.transpose(self._output_transpose), out)

    def flush_buffers(self, constants=None):
        """Optional reset internal buffers (implement in subclasses which need this)"""
        pass

    def _update_correction_flags(self):
        enabled = self._correction_flag_class.NONE
        output = Hash()  # for informing downstream receivers what was applied
        preview = self._correction_flag_class.NONE
        for field_name, flag, constants in self._correction_steps:
            output[field_name] = False
            if self._get(f"corrections.{field_name}.available"):
                if self._get(f"corrections.{field_name}.enable"):
                    enabled |= flag
                    output[field_name] = True
                if self._get(f"corrections.{field_name}.preview"):
                    preview |= flag
        self._correction_flag_enabled = enabled
        self._correction_flag_preview = preview
        self._correction_applied_hash = output
        self._device.log.DEBUG(f"Corrections for dataOutput: {str(enabled)}")
        self._device.log.DEBUG(f"Corrections for preview: {str(preview)}")

    def _update_bad_pixel_subset(self):
        selection = 0
        for field in BadPixelValues:
            if self._get(f"corrections.badPixels.subsetToUse.{field.name}"):
                selection |= field
        self.bad_pixel_subset = selection

    def _get(self, key):
        """Will look up key in host device's parameters.  Note that device calls
        reconfigure during preReconfigure, but with new unmerged config overlaid
        temporarily, making _get get the "new" state of things."""
        return self._device.unsafe_get(key)

    def _set(self, key, value):
        """Helper function to host device's configuration. Note that this does *not*
        call reconfigure, so either do so afterwards or make sure to update auxiliary
        properties correctly yourself."""
        self._device.set(key, value)

    @property
    def _map_shape(self):
        return (self._constant_memory_cells, self.num_pixels_ss, self.num_pixels_fs)

    def load_constant(self, constant, constant_data):
        self._load_constant(constant, constant_data)
        self._post_load_constant(constant)

    def _post_load_constant(self, constant):
        for field_name in self._constant_to_correction_names.get(constant, []):
            key = f"corrections.{field_name}.available"
            if not self._get(key):
                self._set(key, True)

        self._update_correction_flags()
        self._device.log_status_info(f"Done loading {constant.name} to runner")

    def correct(self, image_data, cell_table, *additional_data):
        """Correct image data, providing both full corrected data and previews

        This method relies on the following auxiliary methods which may need to be
        overridden / defined in subclass:

        _make_output_buffers
        _preview_data_views
        _correct (must be implemented by subclass)

        Look at signature for _correct; it is given image_data, cell_table,
        *additional_data (anything the host device class' _get_data_from_hash gives us
        in addition to image data, cell table, and pulse table), and *processed_buffers
        which is the result of _make_output_buffers. So all these four methods must
        agree on the relevant set of buffers for input and output.

        The pulse_table is only used for preview index selection purposes as corrections
        do not (yet, at any rate) take the pulse table into account.
        """
        # TODO: check if this is robust enough (also base_correction)
        num_frames = image_data.shape[0]
        processed_buffers = self._make_output_buffers(
            num_frames, self._correction_flag_preview
        )
        image_data = self._xp.asarray(image_data)  # if cupy, will put on GPU
        if cell_table is not None:
            cell_table = self._xp.asarray(cell_table)
        additional_data = [self._xp.asarray(data) for data in additional_data]
        self._correct(
            self._correction_flag_preview,
            image_data,
            cell_table,
            *additional_data,
            *processed_buffers,
        )

        preview_buffers = self._preview_data_views(
            image_data, *additional_data, *processed_buffers
        )

        if self._correction_flag_preview != self._correction_flag_enabled:
            processed_buffers = self._make_output_buffers(
                num_frames, self._correction_flag_enabled
            )
            self._correct(
                self._correction_flag_enabled,
                image_data,
                cell_table,
                *additional_data,
                *processed_buffers,
            )
        return self._correction_applied_hash, processed_buffers[0], preview_buffers


kernel_dir = pathlib.Path(__file__).absolute().parent / "kernels"


def get_kernel_template(kernel_fn):
    with (kernel_dir / kernel_fn).open("r") as fd:
        return jinja2.Template(fd.read())


class KernelRunnerTypes(enum.Enum):
    GPU = enum.auto()
    CPU = enum.auto()
