import h5py
import numpy as np
from calngUtils import shmem_utils
from karabo.bound import (
    DOUBLE_ELEMENT, INT32_ELEMENT, NDARRAY_ELEMENT, STRING_ELEMENT,
    VECTOR_STRING_ELEMENT)

from .. import base_kernel_runner
from .base_addon import BaseCorrectionAddon


def block(b):
    return np.concatenate([np.concatenate(row, axis=-1) for row in b], axis=-2)


def autocorr2_fft(f, mode='full', backend='scipy'):
    """Computes 2D autocorrelation function for real input.

        This is equivalent to
        fftconvolve(f, np.flip(f, axis=(-2, -1)), mode)
    """
    nx, ny = f.shape[-2:]

    from scipy.fft import next_fast_len
    fshape = [next_fast_len(sz, True) for sz in (2 * nx - 1, 2 * ny - 1)]
    r = fft.rfft2(f, fshape)
    r = fft.irfft2((r * r.conj()).real, fshape)

    if mode == "same":
        ux = (nx + 1) // 2
        lx = (1 - nx) // 2
        uy = (ny + 1) // 2
        ly = (1 - ny) // 2
    elif mode == "full":
        ux = nx
        lx = (1 - nx)
        uy = ny
        ly = (1 - ny)
    else:
        raise ValueError("acceptable mode flags are 'same', or 'full'")

    a11 = r[..., :ux, :uy]
    a12 = r[..., :ux, ly:]
    a21 = r[..., lx:, :uy]
    a22 = r[..., lx:, ly:]

    return block([[a22, a21], [a12, a11]])


def autocorr2symm_fft(f, backend='scipy'):
    """Computes 2D autocorrelation function for real input and
       symmetrical boundary conditions.
    """
    r = fft.dctn(f, 3, axes=(-2, -1))
    r = fft.idctn(r * r, 3, axes=(-2, -1))
    return r


def autocorr2wrap_fft(f, backend='scipy'):
    """Computes 2D autocorrelation function for real input and
       circular boundary conditions.
    """
    s = f.shape[-2:]
    r = fft.rfft2(f, s)
    r = fft.irfft2((r * r.conj()).real, s)
    r = fft.fftshift(r)
    return r


AUTOCORR_FUN = {
    "symm": autocorr2symm_fft,
    "wrap": autocorr2wrap_fft,
    "fill": lambda f: autocorr2_fft(f, "same"),
}

# AGIPD specific, we need to know the detector layout
RESULT_SHAPE = {
    "symm": np.s_[:, :-14, :],
    "wrap": np.s_[:, 7:-7, :],
    "fill": np.s_[:, 7:-7, :],
}


class Autocorrelation(BaseCorrectionAddon):
    _device = None  # will be set to host device *after* init

    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            INT32_ELEMENT(schema)
            .key(f"{prefix}.moduleNumber")
            .readOnly().initialValue(-1)
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.boundary")
            .tags("managed")
            .options(",".join(AUTOCORR_FUN.keys()))
            .assignmentOptional()
            .defaultValue("symm")
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.intenstityPerPhoton")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(9.3)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.roundingThreshold")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0.7)
            .reconfigurable()
            .commit(),

            VECTOR_STRING_ELEMENT(schema)
            .key(f"{prefix}.maskPaths")
            .tags("managed")
            .assignmentOptional()
            .defaultValue([])
            .reconfigurable()
            .commit(),
        )

    @staticmethod
    def extend_output_schema(schema):
        (
            NDARRAY_ELEMENT(schema)
            .key("image.autocorr")
            .dtype('FLOAT')
            .commit(),
        )

    def __init__(self, device, prefix, config):
        # note: this includes reconfigure
        super().__init__(device, prefix, config)
        self._shape = None
        self._shmem_buffer = None

        global cupy
        import cupy

        global fft
        kernel_type = base_kernel_runner.KernelRunnerTypes[
            device.unsafe_get("kernelType")
        ]
        if kernel_type is base_kernel_runner.KernelRunnerTypes.CPU:
            import scipy.fft as fft
        else:
            import cupyx.scipy.fft as fft

        da = device["constantParameters.karaboDa"]
        self._modno = int(da[-2:])
        device.set(f"{prefix}.moduleNumber", self._modno)

        # AGIPD specific, we need to know the detector layout
        self._repeats = np.ones(512, int)
        self._repeats[63:-1:64] = 2
        self._repeats[64::64] = 2
        self._repeats = self._repeats.tolist()

    def __del__(self):
        del self._shmem_buffer
        super().__del__()

    def post_correction(self, train_id, processed_data, cell_table, pulse_table, output_hash):
        shape = processed_data.shape
        if shape != self._shape:
            self._update_buffer(shape)

        # conversion to photons
        np.around(
            processed_data / self._intensity_per_photon - (self._rounding_threshold - 0.5),
            out=processed_data
        )
        processed_data[processed_data < 0.0] = 0.0

        # apply extra user mask
        if self._mask is not None:
            processed_data[:, self._mask] = np.nan

        #  compute autocorrelation
        data = processed_data.copy()
        data[np.isnan(data)] = 0.0
        data = np.repeat(data, self._repeats, axis=1)

        autocorr = self._autocorr(data)[self._result_shape]

        # compute autocorrelation normalized by mask (takes too much memory)
        # mask = np.isfinite(data)
        # mask = np.repeat(mask, self._repeats, axis=1).astype(np.float32)
        # autocorr /= self._autocorr(mask)[self._result_shape]

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        if hasattr(autocorr, 'get'):
            autocorr.get(out=buffer_array[:])
        else:
            buffer_array[:] = autocorr

        if self._device._use_shmem_handles:
            output_hash.set("image.autocorr", buffer_handle)
            shmem_paths = output_hash.get("calngShmemPaths", default=[])
            output_hash.set("calngShmemPaths",
                            shmem_paths + ["image.autocorr"])
        else:
            output_hash.set("image.autocorr", buffer_array)

    def reconfigure(self, changed_config):
        if changed_config.has("intenstityPerPhoton"):
            self._intensity_per_photon = changed_config["intenstityPerPhoton"]
        if changed_config.has("boundary"):
            self._autocorr = AUTOCORR_FUN[changed_config["boundary"]]
            self._result_shape = RESULT_SHAPE[changed_config["boundary"]]
        if changed_config.has("roundingThreshold"):
            self._rounding_threshold = changed_config["roundingThreshold"]
        if changed_config.has("maskPaths"):
            self._load_mask(changed_config["maskPaths"])

    def _update_buffer(self, shape):
        self._shape = shape

        if self._shmem_buffer is None:
            shmem_buffer_name = self._device.getInstanceId() + ":dataOutput/autocorrelation"
            memory_budget = self._device.get("outputShmemBufferSize") * 2**30
            self._shmem_buffer = shmem_utils.ShmemCircularBuffer(
                memory_budget,
                shape,
                np.float32,
                shmem_buffer_name,
            )
            if self._device._cuda_pin_buffers:
                self._shmem_buffer.cuda_pin()
        else:
            self._shmem_buffer.change_shape(shape)

    def _load_mask(self, paths):
        self._mask = None
        for fn in paths:
            try:
                with h5py.File(fn, 'r') as f:
                    mask = f["entry_1/data_1/mask"][self._modno] != 0
                if self._mask is None:
                    self._mask = mask
                else:
                    self._mask = self._mask | mask
            except Exception as e:
                self._device.logger.warn(f"Aurocorrelation addon: {e}")
