import re

from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    IMAGEDATA_ELEMENT,
    INT32_ELEMENT,
    NDARRAY_ELEMENT,
    NODE_ELEMENT,
    UINT32_ELEMENT,
    EventLoop,
    Hash,
    ImageData,
)
import numpy as np

from .base_addon import BaseCorrectionAddon
from .. import azim_utils
from ..utils import maybe_get


class AzimuthalIntegrator(BaseCorrectionAddon):
    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            UINT32_ELEMENT(schema)
            .key(f"{prefix}.numBins")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(1000)
            .reconfigurable()
            .commit(),

            # TODO: automatic decision from integration with device class
            # or enable using GPU even if rest of device didn't
            BOOL_ELEMENT(schema)
            .key(f"{prefix}.useGpu")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key(f"{prefix}.multiModule")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(schema)
            .key(f"{prefix}.moduleIndex")
            .assignmentOptional()
            .defaultValue(-1)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key(f"{prefix}.pixelsplittingFactor")
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.detectorDistance")
            .assignmentOptional()
            .defaultValue(0.5)
            .reconfigurable()
            .commit(),
        )

    @staticmethod
    def extend_output_schema(schema):
        (
            NODE_ELEMENT(schema)
            .key("azimuthal")
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("azimuthal.integrated1d")
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("azimuthal.integrated1dWeight")
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("azimuthal.preview")
            .commit(),

            IMAGEDATA_ELEMENT(schema)
            .key("azimuthal.streak")
            .commit(),
        )

    def __init__(self, device, prefix, conf):
        self._conf = Hash()
        self.integrator = None
        # full is entire detector, local is for this module (if case multi-module det)
        self.full_geom = None
        self.local_geom = None
        super().__init__(device, prefix, conf)
        self._try_to_guess_module_index()

    def reconfigure(self, conf):
        self._conf.merge(conf)
        if self.full_geom is None:
            # cannot do anything below yet
            return
        if any(conf.has(key) for key in ("useGpu", "multiModule", "numBins")):
            self._compute_bins()
            EventLoop.post(self._initialize_integrator)
        elif any(conf.has(key) for key in ("pixelsplittingFactor", "useGpu")):
            EventLoop.post(self._initialize_integrator)

    def on_new_geometry(self, geom):
        self.full_geom = geom
        if self._conf["multiModule"]:
            # TODO: warn if module index not known
            self.local_geom = azim_utils.slice_geom(
                self.full_geom, self._conf["moduleIndex"]
            )
        else:
            self.local_geom = self.full_geom
        self._compute_bins()
        EventLoop.post(self._initialize_integrator)

    def _compute_bins(self):
        if self.full_geom is None:
            self._device.log_status_warn(
                "Azimuthal integration addon cannot compute bins because it doesn't know "
                "its geometry"
            )
            return
        self.bins = np.linspace(
            0,
            azim_utils.get_geom_max_theta(self.full_geom, self._conf["detectorDistance"]),
            self._conf["numBins"],
            dtype=np.float32
        )

    def _initialize_integrator(self):
        self._device.log_status_info("Azimuthal integration addon computing CSR...")
        if self._conf["useGpu"]:
            self.integrator = azim_utils.GpuAzimuthalIntegrator(
                self.local_geom,
                self.bins,
                self._conf["detectorDistance"],
                self._conf["pixelsplittingFactor"],
            )
        else:
            self.integrator = azim_utils.CpuAzimuthalIntegrator(
                self.local_geom,
                self.bins,
                self._conf["detectorDistance"],
                self._conf["pixelsplittingFactor"],
            )
        self._device.log_status_info("Azimuthal integration addon CSR is ready")

    def post_correction(self, train_id, processed_data, cell_table, pulse_table, output_hash):
        if self.integrator is None:
            self._device.log_status_warn("Azimuthal integration addon not ready yet")
            return
        processed_data = processed_data.reshape(processed_data.shape[0], -1)
        if self._conf["multiModule"]:
            data, weight = self.integrator.integrate_1d_multimodule(processed_data)
            output_hash["azimuthal.integrated1d"] = maybe_get(data)
            output_hash["azimuthal.integrated1dWeight"] = maybe_get(weight)
            # TODO: don't assume multi-frame data
            previewable = data / weight
        else:
            data, weight = self.integrator.integrate_1d(processed_data)
            output_hash["azimuthal.integrated1d"] = maybe_get(data)
            previewable = data
        output_hash["azimuthal.bins"] = self.bins
        output_hash["azimuthal.preview"] = maybe_get(
            self.integrator._xp.nanmean(previewable, axis=0)
        )
        output_hash["azimuthal.streak"] = ImageData(
            maybe_get(
                self.integrator._xp.nan_to_num(previewable)
            )
        )

    def _try_to_guess_module_index(self):
        for source in self._device.get("fastSources"):
            # stolen from masking addon
            for (source_pattern, extractor) in [
                (
                    r".*\/DET\/(\d+)CH0:xtdf",  # XTDF-detector source
                    lambda match: int(match.group(1)),
                ),
                (
                    r".*\/DET\/.*?(\d+):daqOutput",  # typical JF receiver source
                    lambda match: int(match.group(1)) - 1,
                    # don't subtract 1; JUNGFRAUGeometry does this for us
                )
            ]:
                if (match := re.match(source_pattern, source)) is not None:
                    module_index = extractor(match)
                    self._device.log_status_info(
                        f"Source {source} means {module_index}"
                    )
                    self._device.set(f"{self._prefix}.moduleIndex", module_index)
                    self._conf["moduleIndex"] = module_index
                    return
        self._device.log_status_info("User mask addon couldn't guess module index")
