import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_INT32_ELEMENT,
    Epochstamp,
    Hash,
    Timestamp,
    Trainstamp,
)

from ..utils import maybe_get
from .base_addon import BaseCorrectionAddon


class RoiIntegration(BaseCorrectionAddon):
    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            STRING_ELEMENT(schema)
            .key(f'{prefix}.pulses')
            .description(
                'Slice pulses to select in the train. Selected pulses '
                'are summed before performing the integration')
            .assignmentOptional()
            .defaultValue('0:1')
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.valueMin")
            .assignmentOptional()
            .defaultValue(-10.0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.valueMax")
            .assignmentOptional()
            .defaultValue(20000.0)
            .reconfigurable()
            .commit(),

            VECTOR_INT32_ELEMENT(schema)
            .key(f"{prefix}.roi")
            .assignmentOptional()
            .defaultValue([0, 10, 0, 10])
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.meanIntensity")
            .readOnly()
            .initialValue(0.)
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.meanIntensityROI")
            .readOnly()
            .initialValue(0.)
            .commit(),
        )

    def reconfigure(self, conf):
        if conf.has("roi"):
            self.x_min, self.x_max, self.y_min, self.y_max = conf["roi"]
        if conf.has("pulses"):
            self.pulses = eval(f'np.s_[{conf["pulses"]}]')
        if conf.has("valueMax"):
            self._vmax = conf["valueMax"]
        if conf.has("valueMin"):
            self._vmin = conf["valueMin"]

    def post_correction(self, train_id, data, cell_table, pulse_table, data_hash):
        data = data[self.pulses]
        mask = np.isfinite(data) & (self._vmin < data) & (data < self._vmax)
        count = np.sum(mask)
        count_roi = np.sum(mask[:, self.y_min:self.y_max, self.x_min:self.x_max])

        masked_data = np.nansum(data * mask, axis=0)
        mean_data = np.nansum(masked_data) / count

        roi = masked_data[self.y_min:self.y_max, self.x_min:self.x_max]
        mean_roi = np.nansum(roi) / count_roi

        res = Hash(
            f"{self._prefix}.meanIntensity", float(maybe_get(mean_data)),
            f"{self._prefix}.meanIntensityROI", float(maybe_get(mean_roi)),
        )

        ts = Timestamp(Epochstamp(), Trainstamp(train_id))
        ts.toHashAttributes(res.getAttributes(f"{self._prefix}.meanIntensity"))
        ts.toHashAttributes(res.getAttributes(f"{self._prefix}.meanIntensityROI"))
        self._device.set(res)
