import re

import numpy as np
import h5py
from karabo.bound import (
    BOOL_ELEMENT,
    INT32_ELEMENT,
    STRING_ELEMENT,
    SLOT_ELEMENT,
)
from .base_addon import BaseCorrectionAddon


class UserMask(BaseCorrectionAddon):
    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            STRING_ELEMENT(schema)
            .key(f"{prefix}.maskFilePath")
            .displayedName("File path")
            .description(
                "Path to an HDF5 file containing mask data. If set, will automatically "
                "try to load on init. Will also try to (re)load when path is changed."
            )
            .tags("managed")
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.dataPath")
            .displayedName("Dataset name")
            .description(
                "Path within the file to go to for the mask data. If left empty, I "
                "will just take the first array I find and try my best. Will reload if "
                "this is set / changed."
            )
            .tags("managed")
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            SLOT_ELEMENT(schema)
            .key(f"{prefix}.loadMask" )
            .displayedName("Load mask")
            .description(
                "(Re)loads the mask from currently specified file + dataset name"
            )
            .tags("managed")
            .commit(),

            BOOL_ELEMENT(schema)
            .key(f"{prefix}.multiModule")
            .displayedName("File covers multiple modules?")
            .description(
                "For multi-module detectors, you probably want to generate a single "
                "mask file rather than many. If this flag is on, the first axis of the "
                "loaded dataset is assumed to correspond to module index. So shape will "
                "be ex. (16, 256, 256) for LPD1M if this is set and burstMode is not "
                "set."
            )
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key(f"{prefix}.burstMode")
            .displayedName("File has per-cell mask?")
            .description(
                "For burst mode detectors, it can make sense to mask based on pixel "
                "*and* memory cell. If this flag is set, the second axis (first axis "
                "in case multiModule is False) is assumed to correspond to memory "
                "cells. Please make sure the cell axis is large enough to accomodate "
                "the memory cells in use. Leave this setting False to just provide a "
                "simple per-pixel mask (definitely the case when masking by hand)."
            )
            .tags("managed")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(schema)
            .key(f"{prefix}.moduleIndex")
            .displayedName("Module index")
            .description(
                "For multi-module detectors, if multiModule is on, we need to know the "
                "correction device's module's index. This is hopefully automatically "
                "set."
            )
            .assignmentOptional()
            .defaultValue(-1)
            .commit(),
        )

    def __init__(self, device, prefix, conf):
        self._module_index = None
        self._mask = None
        super().__init__(device, prefix, conf)
        self._device.KARABO_SLOT(
            self._load_mask,
            slotName=f"{prefix}.loadMask".replace(".", "_"),
            numArgs=0,
        )

    def _try_to_guess_module_index(self):
        for source in self._device.get("fastSources"):
            # stolen from DetectorAssembler
            # TODO: something configurable, robust
            for (source_pattern, extractor) in [
                (
                    r".*\/DET\/(\d+)CH0:xtdf",  # XTDF-detector source
                    lambda match: int(match.group(1)),
                ),
                (
                    r".*\/DET\/.*?(\d+):daqOutput",  # typical JF receiver source
                    lambda match: int(match.group(1)) - 1,
                    # don't subtract 1; JUNGFRAUGeometry does this for us
                )
            ]:
                if (match := re.match(source_pattern, source)) is not None:
                    self._module_index = extractor(match)
                    self._device.log_status_info(
                        f"Source {source} means {self._module_index}"
                    )
                    return
        self._device.log_status_info("User mask addon couldn't guess module index")

    def reconfigure(self, conf):
        if conf.has("moduleIndex"):
            if conf["moduleIndex"] < 0:
                self._try_to_guess_module_index()
                if self._module_index is not None:
                    conf["moduleIndex"] = self._module_index
            else:
                self._module_index = conf["moduleIndex"]
        if conf.has("multiModule"):
            if conf["multiModule"] and self._module_index is None:
                warning = "Unable to guess module index, cannot use multi-module mask files"
                # only throw if we're /not/ in device init
                if hasattr(self, "_past_init"):
                    # but interrupt bad preReconfigure with exception
                    raise ValueError(warning)
                else:
                    self._device.log_status_warn(f"{self._prefix}: {warning}")
            self._multi_module = conf["multiModule"]
        if conf.has("burstMode"):
            self._burst_mode = conf["burstMode"]
        if conf.has("maskFilePath"):
            self._mask_fn = conf["maskFilePath"]
        if conf.has("dataPath"):
            self._mask_data_path = conf["dataPath"]
            self._user_mask_data_path = self._mask_data_path
        if self._mask_fn and (conf.has("maskFilePath") or conf.has("dataPath")):
            try:
                self._load_mask()
            except Exception as ex:
                self._device.log_status_warn(
                    f"{self._prefix} failed to load mask: {ex}"
                )
                if not initial:
                    raise ex
        self._past_init = True


    def _load_mask(self):
        """Utility function to load mask from currently set path and such. Can throw
        all kinds of exceptions, caller should decide whether to catch them."""
        with h5py.File(self._mask_fn, "r") as fd:
            if not self._user_mask_data_path:
                self._guess_data_path(fd)
            mask = np.array(fd[self._mask_data_path])
        if self._multi_module:
            assert (
                self._module_index is not None
            ), "Tried loading multi module, but I don't know my module index"
            assert (
                self._module_index <= mask.shape[0]
            ), "My module index is out of bounds for this mask file"
            mask = mask[self._module_index]
        if self._burst_mode:
            assert mask.ndim == 3, f"Burst mode but mask has {mask.ndim} dimensions"
        else:
            assert mask.ndim == 2, f"Single-frame mode but mask has {mask.ndim} dimensions"
        self._mask = (mask != 0)

    def _guess_data_path(self, fd):
        """People probably don't want to specify the data path, but if the mask is the
        only Dataset in the file, then we can just try to guess it."""
        def visitor(key, val):
            if isinstance(val, h5py.Dataset):
                self._device.log_status_info(
                    f"User mask addon assumes mask path is: {key}"
                )
                self._mask_data_path = key
                self._device.set(f"{self._prefix}.dataPath", self._mask_data_path)
                raise StopIteration()
        try:
            fd.visititems(visitor)
        except StopIteration:
            pass
        else:
            raise ValueError(f"No dataset found in {fd.filename}")

    def post_correction(self, train_id, data, cell_table, pulse_table, output_hash):
        if self._burst_mode:
            try:
                mask = self._mask[cell_table]
            except IndexError as ex:
                self._device.log_status_warn(
                    f"Failed to get mask in cell order (mask too small?), {ex}"
                )
                return
        else:
            mask = self._mask
        data[..., mask] = np.nan
