import numpy as np

from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    IMAGEDATA_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    Dims,
    Encoding,
    Epochstamp,
    Hash,
    ImageData,
    Schema,
    Timestamp,
    Trainstamp,
)

from .base_addon import BaseCorrectionAddon


def maybe_get(a):
    # TODO: proper check for cupy
    if hasattr(a, "get"):
        return a.get()
    return a


def saturation_monitoring_schema(schema=None):
    if schema is None:
        schema = Schema()
    (
        NODE_ELEMENT(schema)
        .key("saturationMonitor")
        .commit(),

        BOOL_ELEMENT(schema)
        .key("saturationMonitor.warning")
        .readOnly()
        .initialValue(False)
        .commit(),

        BOOL_ELEMENT(schema)
        .key("saturationMonitor.alarm")
        .readOnly()
        .initialValue(False)
        .commit(),

        UINT32_ELEMENT(schema)
        .key("saturationMonitor.warnCount")
        .description(
            "Total number of pixels above warning threshold. Each pixel "
            "is only counted once even if it exceeds the threshold in "
            "multiple frames (/ memory cells = given axis)."
        )
        .readOnly()
        .initialValue(0)
        .commit(),

        UINT32_ELEMENT(schema)
        .key("saturationMonitor.alarmCount")
        .description(
            "Total number of pixels above alarm threshold. Each pixel "
            "is only counted once even if it exceeds the threshold in "
            "multiple frames (/ memory cells = given axis)."
        )
        .readOnly()
        .initialValue(0)
        .commit(),

        FLOAT_ELEMENT(schema)
        .key("saturationMonitor.maxValue")
        .description("Max pixel value in latest train with warning or alarm.")
        .readOnly()
        .initialValue(0)
        .commit(),

        UINT64_ELEMENT(schema)
        .key("saturationMonitor.trainId")
        .readOnly()
        .initialValue(0)
        .commit(),
    )
    return schema


class SaturationMonitor(BaseCorrectionAddon):
    def __init__(self, device, prefix, config):
        super().__init__(device, prefix, config)
        global cupy
        import cupy
        self._alarmThreshold = config["alarmThreshold"]
        self._warnThreshold = config["warnThreshold"]
        self._alarmMaxCount = config["alarmMaxCount"]
        self._warnMaxCount = config["warnMaxCount"]
        self._frameAxis = config["frameAxis"]

    def reconfigure(self, changed_config):
        if changed_config.has("alarmThreshold"):
            self._alarmThreshold = changed_config["alarmThreshold"]
        if changed_config.has("warnThreshold"):
            self._warnThreshold = changed_config["warnThreshold"]
        if changed_config.has("alarmMaxCount"):
            self._alarmMaxCount = changed_config["alarmMaxCount"]
        if changed_config.has("warnMaxCount"):
            self._warnMaxCount = changed_config["warnMaxCount"]
        if changed_config.has("frameAxis"):
            self._frameAxis = changed_config["frameAxis"]

    @classmethod
    def extend_device_schema(cls, schema, prefix):
        # will add own output channel with this schema
        output_schema = saturation_monitoring_schema()
        (
            IMAGEDATA_ELEMENT(output_schema)
            .key("saturationMonitor.alarmImage")
            .commit(),
        )
        (
            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.alarmThreshold")
            .description("Alarm threshold per pixel.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.warnThreshold")
            .description("Warning threshold per pixel.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            UINT64_ELEMENT(schema)
            .key(f"{prefix}.alarmMaxCount")
            .description("Maximum number of pixel above alarm threshold.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            UINT64_ELEMENT(schema)
            .key(f"{prefix}.warnMaxCount")
            .description("Maximum number of pixel above warn threshold.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            UINT64_ELEMENT(schema)
            .key(f"{prefix}.frameAxis")
            .displayedName("Multi-frame axis")
            .description("Axis for frames. Used to take the max over this axis.")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            OUTPUT_CHANNEL(schema)
            .key(f"{prefix}.output")
            .dataSchema(output_schema)
            .commit(),
        )

    def post_correction(self, train_id, data, cell_table, pulse_table, data_hash):
        output = Hash()

        # only take the max if data has frames -> more than 2 dimensions
        if data.ndim > 2:
            max_image = np.nanmax(data, axis=self._frameAxis)
        else:
            max_image = data
        nb_pix_warning = int(np.nansum(max_image > self._warnThreshold))
        nb_pix_alarm = int(np.nansum(max_image > self._alarmThreshold))

        output["saturationMonitor.warning"] = nb_pix_warning > self._warnMaxCount
        output["saturationMonitor.alarm"] = nb_pix_alarm > self._alarmMaxCount
        output["saturationMonitor.warnCount"] = nb_pix_warning
        output["saturationMonitor.alarmCount"] = nb_pix_alarm
        output["saturationMonitor.maxValue"] = float(np.nanmax(max_image))
        output["saturationMonitor.trainId"] = train_id
        max_image[max_image <= self._warnThreshold] = 0
        output["saturationMonitor.alarmImage"] = ImageData(
            maybe_get(max_image), Dims(*max_image.shape), Encoding.GRAY, bitsPerPixel=32
        )
        self._device.writeChannel(
            f"{self._prefix}.output",
            output,
            timestamp=Timestamp(Epochstamp(), Trainstamp(train_id)),
        )
