import concurrent.futures
import enum
import functools

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    MetricPrefix,
    Unit,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    utils,
)
from ..preview_utils import PreviewFriend, FrameSelectionMode, PreviewSpec
from .._version import version as deviceVersion


class Constants(enum.Enum):
    Offset10Hz = enum.auto()
    BadPixelsDark10Hz = enum.auto()
    BadPixelsFF10Hz = enum.auto()
    RelativeGain10Hz = enum.auto()


bad_pixel_constants = {
    Constants.BadPixelsDark10Hz,
    Constants.BadPixelsFF10Hz,
}


# from pycalibration (TOOD: move to common shared lib)
class GainMode(utils.OptionsEnum):
    ADAPTIVE_GAIN = 0
    FIXED_GAIN = 1


class GainSetting(utils.OptionsEnum):
    LOW_CDS = 0
    HIGH_CDS = 1


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1
    REL_GAIN = 2
    BPMASK = 4
    STRIXEL = 8


correction_steps = (
    ("offset", CorrectionFlags.OFFSET, {Constants.Offset10Hz}),
    ("relGain", CorrectionFlags.REL_GAIN, {Constants.RelativeGain10Hz}),
    (
        "badPixels",
        CorrectionFlags.BPMASK,
        {
            Constants.BadPixelsDark10Hz,
            Constants.BadPixelsFF10Hz,
            None,
        },
    ),
    ("strixel", CorrectionFlags.STRIXEL, set()),
)


class JungfrauBaseRunner(base_kernel_runner.BaseKernelRunner):
    _bad_pixel_constants = bad_pixel_constants
    _correction_flag_class = CorrectionFlags
    _correction_steps = correction_steps
    num_pixels_ss = 512
    num_pixels_fs = 1024

    @property
    def _map_shape(self):
        return (self._constant_memory_cells, self.num_pixels_ss, self.num_pixels_fs, 3)

    def _setup_constant_buffers(self):
        # note: superclass creates cell table with wrong dtype
        self.offset_map = self._xp.empty(self._map_shape, dtype=np.float32)
        self.rel_gain_map = self._xp.empty(self._map_shape, dtype=np.float32)
        self.bad_pixel_map = self._xp.empty(self._map_shape, dtype=np.uint32)
        self.flush_buffers(set(Constants))

    def _make_output_buffers(self, num_frames, flags):
        if flags & CorrectionFlags.STRIXEL:
            return [
                self._xp.empty(
                    (num_frames,) + self._strixel_out_shape, dtype=self._output_dtype
                )
            ]
        else:
            return [
                self._xp.empty(
                    (num_frames, self.num_pixels_ss, self.num_pixels_fs),
                    dtype=self._output_dtype,
                )
            ]

    def _expected_output_shape(self, num_frames):
        # TODO: think of how to unify with _make_output_buffers
        if self._correction_flag_enabled & CorrectionFlags.STRIXEL:
            return (num_frames,) + self._strixel_out_shape
        else:
            return (num_frames, self.num_pixels_ss, self.num_pixels_fs)

    @classmethod
    def add_schema(cls, schema):
        super(cls, cls).add_schema(schema)
        super(cls, cls).add_bad_pixel_config(schema)
        (
            OVERWRITE_ELEMENT(schema)
            .key("corrections.strixel.enable")
            .setNewDefaultValue(False)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("corrections.strixel.preview")
            .setNewDefaultValue(False)
            .commit(),

            STRING_ELEMENT(schema)
            .key("corrections.strixel.type")
            .description(
                "Which kind of strixel layout is used for this module? cols_A0123 is "
                "the first strixel layout deployed at HED and rows_A1256 is the one "
                "later deployed at SCS."
            )
            .assignmentOptional()
            .defaultValue("cols_A0123(HED-type)")
            .options("cols_A0123(HED-type),rows_A1256(SCS-type)")
            .reconfigurable()
            .commit(),
        )

    def reconfigure(self, config):
        super().reconfigure(config)

        if (strixel_type := config.get("corrections.strixel.type")) is not None:
            # drop the friendly parenthesized name
            strixel_type = strixel_type.partition("(")[0]
            strixel_package = np.load(
                base_kernel_runner.kernel_dir
                / f"strixel_{strixel_type}-lut_mask.npz"
            )
            self._strixel_out_shape = tuple(strixel_package["frame_shape"])
            # TODO: use bad pixel masking config here
            self.correction_kernel_strixel = functools.partial(
                utils.apply_partial_lut,
                lut=self._xp.asarray(strixel_package["lut"]),
                mask=self._xp.asarray(strixel_package["mask"]),
                missing=self.bad_pixel_mask_value,
            )
            # note: always masking unmapped pixels (not respecting NON_STANDARD_SIZE)

    def _load_constant(self, constant, constant_data):
        if constant_data.shape[0] == self.num_pixels_fs:
            constant_data = np.transpose(constant_data, (2, 1, 0, 3))
        else:
            constant_data = np.transpose(constant_data, (2, 0, 1, 3))

        if constant is Constants.Offset10Hz:
            self.offset_map[:] = self._xp.asarray(constant_data, dtype=np.float32)
        elif constant is Constants.RelativeGain10Hz:
            self.rel_gain_map[:] = self._xp.asarray(constant_data, dtype=np.float32)
        elif constant in bad_pixel_constants:
            self.bad_pixel_map |= self._xp.asarray(
                constant_data, dtype=np.uint32
            ) & self._xp.uint32(self.bad_pixel_subset)
        else:
            raise ValueError(f"Unexpected constant type {constant}")

    def _preview_data_views(self, raw_data, gain_map, processed_data):
        return (raw_data, gain_map, processed_data)

    def flush_buffers(self, constants):
        if Constants.Offset10Hz in constants:
            self.offset_map.fill(0)
        if Constants.RelativeGain10Hz in constants:
            self.rel_gain_map.fill(1)
        if constants & {
            Constants.BadPixelsDark10Hz,
            Constants.BadPixelsFF10Hz,
        }:
            self.bad_pixel_map.fill(0)
            if self.bad_pixel_subset & utils.BadPixelValues.NON_STANDARD_SIZE:
                self._mask_asic_seams()

    def _mask_asic_seams(self):
        self.bad_pixel_map[
            :, :, 255:1023:256
        ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
        self.bad_pixel_map[
            :, :, 256:1024:256
        ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
        self.bad_pixel_map[
            :, [255, 256]
        ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value


class JungfrauGpuRunner(JungfrauBaseRunner):
    runner_type = base_kernel_runner.KernelRunnerTypes.GPU

    def _post_init(self):
        source_module = self._xp.RawModule(
            code=base_kernel_runner.get_kernel_template("jungfrau_gpu.cu").render(
                {
                    "output_data_dtype": utils.np_dtype_to_c_type(
                        self._output_dtype
                    ),
                    "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                }
            )
        )
        self.correction_kernel = source_module.get_function("correct")

    def _correct(self, flags, image_data, cell_table, gain_map, output):
        num_frames = image_data.shape[0]
        block_shape = (1, 1, 64)
        grid_shape = utils.grid_to_cover_shape_with_blocks(
            output.shape, block_shape
        )
        if flags & CorrectionFlags.STRIXEL:
            # do "regular" correction into "regular" buffer, strixel transform to output
            final_output = output
            output = self._xp.empty_like(image_data, dtype=np.float32)
        self.correction_kernel(
            grid_shape,
            block_shape,
            (
                image_data,
                gain_map,
                cell_table,
                self._xp.uint8(flags),
                self._xp.uint16(num_frames),
                self._xp.uint16(self._constant_memory_cells),
                self._xp.uint8(num_frames > 1),
                self.offset_map,
                self.rel_gain_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                output,
            ),
        )
        if flags & CorrectionFlags.STRIXEL:
            for pixel_frame, strixel_frame in zip(output, final_output):
                self.correction_kernel_strixel(
                    data=pixel_frame,
                    out=strixel_frame,
                )


class JungfrauCpuRunner(JungfrauBaseRunner):
    runner_type = base_kernel_runner.KernelRunnerTypes.CPU

    def _post_init(self):
        # for computing previews faster
        self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=16)

        from ..kernels import jungfrau_cython

        self.correction_kernel_single = jungfrau_cython.correct_single
        self.correction_kernel_burst = jungfrau_cython.correct_burst

    def __del__(self):
        self.thread_pool.shutdown()

    def _correct(self, flags, image_data, cell_table, gain_map, output):
        if flags & CorrectionFlags.STRIXEL:
            # do "regular" correction into "regular" buffer, strixel transform to output
            final_output = output
            output = self._xp.empty_like(image_data, dtype=np.float32)

        # burst mode
        if image_data.shape[0] > 1:
            self.correction_kernel_burst(
                image_data,
                gain_map,
                cell_table,
                flags,
                self.offset_map,
                self.rel_gain_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                output,
            )
        else:
            self.correction_kernel_single(
                image_data,
                gain_map,
                flags,
                self.offset_map,
                self.rel_gain_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                output,
            )

        if flags & CorrectionFlags.STRIXEL:
            # now do strixel transformation into actual output buffer
            concurrent.futures.wait(
                [
                    self.thread_pool.submit(
                        self.correction_kernel_strixel,
                        data=pixel_frame,
                        out=strixel_frame,
                    )
                    for pixel_frame, strixel_frame in zip(output, final_output)
                ]
            )


class JungfrauCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {
            Constants.Offset10Hz: self.dark_condition,
            Constants.BadPixelsDark10Hz: self.dark_condition,
            Constants.BadPixelsFF10Hz: self.base_condition,
            Constants.RelativeGain10Hz: self.base_condition,
        }

    @staticmethod
    def add_schema(schema):
        super(JungfrauCalcatFriend, JungfrauCalcatFriend).add_schema(
            schema, "jungfrau-Type"
        )

        # set some defaults for common parameters
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(1024)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(90)
            .commit(),
        )

        # add extra parameters
        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.integrationTime")
            .tags("managed")
            .displayedName("Integration time")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MICRO)
            .description(
                "Note that the control device specifies integration time in seconds "
                "and we here expect microseconds. This is because injected constant "
                "conditions use microseconds - despite CalCat thinking the parameter "
                "is in seconds."
            )
            .assignmentOptional()
            .defaultValue(350)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.exposureTimeout")
            .tags("managed")
            .displayedName("Exposure timeout")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.NANO)
            .assignmentOptional()
            .defaultValue(25)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.sensorTemperature")
            .tags("managed")
            .displayedName("Sensor temperature")
            .unit(Unit.KELVIN)
            .description("Sensor temperature in K")
            .assignmentOptional()
            .defaultValue(291)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.gainMode")
            .tags("managed")
            .displayedName("Gain mode")
            .description(
                "Detector may be operating in one of several gain modes. In CalCat, "
                "these map to two CalCat parameters: 'Gain Setting' which is 1 for "
                "dynamic gain with HG0 (0 otherwise) and 'Gain mode' which is 1 "
                "for fixed gain (omitted otherwise)."
            )
            .unit(Unit.NUMBER)
            .assignmentOptional()
            .defaultValue(GainMode.ADAPTIVE_GAIN.format_helpful())
            .options(GainMode.list_options())
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.gainSetting")
            .tags("managed")
            .displayedName("Gain setting")
            .description("See description of gainMode")
            .assignmentOptional()
            .defaultValue(GainSetting.LOW_CDS.format_helpful())
            .options(GainSetting.list_options())
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def base_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Integration Time"] = self._get_param("integrationTime")
        res["Sensor Temperature"] = self._get_param("sensorTemperature")

        if (
            gain_mode := GainMode.parse_helpful_format(self._get_param("gainMode"))
        ) is not GainMode.ADAPTIVE_GAIN:
            # NOTE: currently only including if parameter for CalCat is 1
            # change if conditions are tidied up in the database
            res["Gain mode"] = gain_mode.value

        res["Gain Setting"] = GainSetting.parse_helpful_format(
            self._get_param("gainSetting")
        ).value

        return res

    def dark_condition(self):
        res = self.base_condition()

        if (exposure_timeout := self._get_param("exposureTimeout")) != 25:
            res["Exposure timeout"] = exposure_timeout

        return res


@KARABO_CLASSINFO("JungfrauCorrection", deviceVersion)
class JungfrauCorrection(base_correction.BaseCorrection):
    _base_output_schema = schemas.jf_output_schema
    _correction_steps = correction_steps
    _calcat_friend_class = JungfrauCalcatFriend
    _constant_enum_class = Constants

    _image_data_path = "data.adc"
    _cell_table_path = "data.memoryCell"
    _pulse_table_path = None

    _preview_outputs = [
        PreviewSpec(
            name,
            dimensions=2,
            frame_reduction=True,
            valid_frame_selection_modes=(
                FrameSelectionMode.FRAME,
                FrameSelectionMode.CELL,
            ),
        ) for name in ("raw", "gain", "corrected")
    ]
    _warn_memory_cell_range = False
    _use_shmem_handles = False

    @classmethod
    def expectedParameters(cls, expected):
        cls._calcat_friend_class.add_schema(expected)
        JungfrauBaseRunner.add_schema(expected)
        base_correction.add_addon_nodes(expected, cls)
        PreviewFriend.add_schema(expected, cls._preview_outputs)
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(
                cls._base_output_schema(use_shmem_handles=cls._use_shmem_handles)
            )
            .commit(),

            # JUNGFRAU data is small, can fit plenty of trains in here
            OVERWRITE_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .setNewDefaultValue(2)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("useShmemHandles")
            .setNewDefaultValue(cls._use_shmem_handles)
            .commit(),
        )
        (
            # support both CPU and GPU kernels
            STRING_ELEMENT(expected)
            .key("kernelType")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(base_kernel_runner.KernelRunnerTypes.GPU.name)
            .options(
                ",".join(
                    kernel_type.name
                    for kernel_type in base_kernel_runner.KernelRunnerTypes
                )
            )
            .reconfigurable()
            .commit(),
        )

    def _get_data_from_hash(self, data_hash):
        image_data = data_hash.get(self._image_data_path)
        gain_data = np.asarray(data_hash.get("data.gain")).astype(np.uint8, copy=False)
        # explicit np.asarray because types are special here
        cell_table = np.asarray(
            data_hash[self._cell_table_path]
        ).astype(np.uint8, copy=False)
        num_frames = cell_table.size
        if self.unsafe_get("workarounds.overrideInputAxisOrder"):
            expected_shape = self.kernel_runner.expected_input_shape(
                num_frames
            )
            if expected_shape != image_data.shape:
                image_data.shape = expected_shape
                gain_data.shape = expected_shape

        return (
            num_frames,
            image_data,  # not as ndarray as runner can do that
            cell_table,
            None,
            gain_data,
        )

    @property
    def output_data_shape(self):
        if self._correction_flag_enabled & CorrectionFlags.STRIXEL:
            axis_lengths = {
                "x": 3090,
                "y": 86,
                "f": self.unsafe_get("dataFormat.filteredFrames"),
            }
            return tuple(
                axis_lengths[axis]
                for axis in self.unsafe_get("dataFormat.outputAxisOrder")
            )
        return super().output_data_shape

    @property
    def _kernel_runner_class(self):
        kernel_type = base_kernel_runner.KernelRunnerTypes[
            self.unsafe_get("kernelType")
        ]
        if kernel_type is base_kernel_runner.KernelRunnerTypes.CPU:
            return JungfrauCpuRunner
        else:
            return JungfrauGpuRunner
