import enum

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    MetricPrefix,
    Unit,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    utils,
)
from ..preview_utils import PreviewFriend, PreviewSpec
from .._version import version as deviceVersion


class Constants(enum.Enum):
    Offset = enum.auto()
    GainAmpMap = enum.auto()
    RelativeGain = enum.auto()
    FFMap = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsFF = enum.auto()
    Noise = enum.auto()


bad_pixel_constants = {
    Constants.BadPixelsDark,
    Constants.BadPixelsFF,
}


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1
    GAIN_AMP = 2
    REL_GAIN = 4
    FF_CORR = 8
    BPMASK = 16
    PARALLELGAIN = 32


class ParallelGainOptions(enum.Enum):
    HIGH_GAIN = 0
    MEDIUM_GAIN = 1
    LOW_GAIN = 2
    THRESHOLD = enum.auto()


correction_steps = (
    ("offset", CorrectionFlags.OFFSET, {Constants.Offset}),
    ("gainAmp", CorrectionFlags.GAIN_AMP, {Constants.GainAmpMap}),
    ("relGain", CorrectionFlags.REL_GAIN, {Constants.RelativeGain}),
    ("flatfield", CorrectionFlags.FF_CORR, {Constants.FFMap}),
    ("badPixels", CorrectionFlags.BPMASK, bad_pixel_constants),
    # for threshold, will need Offset and Noise
    # manual selection of one always works
    ("parallelGain", CorrectionFlags.PARALLELGAIN, {None}),
)


class LpdBaseRunner(base_kernel_runner.BaseKernelRunner):
    _bad_pixel_constants = bad_pixel_constants
    _correction_flag_class = CorrectionFlags
    _correction_steps = correction_steps
    num_pixels_ss = 256
    num_pixels_fs = 256

    @classmethod
    def add_schema(cls, schema):
        super(cls, cls).add_schema(schema)
        super(cls, cls).add_bad_pixel_config(schema)

        (
            OVERWRITE_ELEMENT(schema)
            .key("corrections.parallelGain")
            .setNewDescription(
                "The processing of parallel gain data is set automatically based on "
                "constantParameters.parallelGain. This node is only used to set the "
                "threshold values to use."
            )
            .commit(),

            STRING_ELEMENT(schema)
            .key("corrections.parallelGain.handling")
            .tags("managed")
            .options(",".join(o.name for o in ParallelGainOptions))
            .assignmentOptional()
            .defaultValue(ParallelGainOptions.THRESHOLD.name)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key("corrections.parallelGain.sigmaHighMedium")
            .tags("managed")
            .description("TODO")
            .assignmentOptional()
            # TODO: good defaults
            .defaultValue(100)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key("corrections.parallelGain.sigmaMediumLow")
            .tags("managed")
            .description("See description of thresholdHighMedium")
            .assignmentOptional()
            # TODO: good defaults
            .defaultValue(200)
            .reconfigurable()
            .commit(),
        )

    def expected_input_shape(self, num_frames):
        return (num_frames, 1, self.num_pixels_ss, self.num_pixels_fs)

    @property
    def _gm_map_shape(self):
        return self._map_shape + (3,)  # for gain-mapped constants

    def _setup_constant_buffers(self):
        self.offset_map = self._xp.empty(self._gm_map_shape, dtype=np.float32)
        self.noise_map = self._xp.empty(self._gm_map_shape, dtype=np.float32)
        self.gain_amp_map = self._xp.empty(self._gm_map_shape, dtype=np.float32)
        self.rel_gain_slopes_map = self._xp.empty(self._gm_map_shape, dtype=np.float32)
        self.flatfield_map = self._xp.empty(self._gm_map_shape, dtype=np.float32)
        self.bad_pixel_map = self._xp.empty(self._gm_map_shape, dtype=np.uint32)
        self.flush_buffers(set(Constants))

    def _make_output_buffers(self, num_frames, flags):
        if flags & CorrectionFlags.PARALLELGAIN:
            num_frames = num_frames // 3
        output_shape = (num_frames, self.num_pixels_ss, self.num_pixels_fs)
        return [
            self._xp.empty(output_shape, dtype=self._xp.float32),  # image
            self._xp.empty(output_shape, dtype=self._xp.float32),  # gain
        ]

    def _expected_output_shape(self, num_frames):
        if self._correction_flag_enabled & CorrectionFlags.PARALLELGAIN:
            num_frames = num_frames // 3
        return super()._expected_output_shape(num_frames)

    def _preview_data_views(self, raw_data, processed_data, gain_map):
        # TODO: always split off gain from raw to avoid messing up preview?
        return (
            raw_data[:, 0],  # raw
            processed_data,  # corrected
            gain_map,  # gain (split from raw)
        )

    def _load_constant(self, constant_type, constant_data):
        # constant type → transpose order, GPU buffer
        transpose_order, my_buffer = {
            Constants.BadPixelsDark: ((2, 1, 0, 3), self.bad_pixel_map),
            Constants.BadPixelsFF: ((2, 0, 1, 3), self.bad_pixel_map),
            Constants.Offset: ((2, 1, 0, 3), self.offset_map),
            Constants.Noise: ((2, 1, 0, 3), self.noise_map),
            Constants.GainAmpMap: ((2, 0, 1, 3), self.gain_amp_map),
            Constants.FFMap: ((2, 0, 1, 3), self.flatfield_map),
            Constants.RelativeGain: ((2, 0, 1, 3), self.rel_gain_slopes_map),
        }[constant_type]

        if constant_type in bad_pixel_constants:
            my_buffer |= (
                self._xp.asarray(
                    constant_data.transpose(transpose_order),
                    dtype=np.uint32,
                )[:self._constant_memory_cells]
                & self._xp.uint32(self.bad_pixel_subset)
            )
        else:
            my_buffer[:] = (
                self._xp.asarray(
                    self._xp.transpose(
                        constant_data.astype(np.float32),
                        transpose_order,
                    )[:self._constant_memory_cells]
                )
            )

    def reconfigure(self, config):
        super().reconfigure(config)
        if config.has("constantParameters.parallelGain"):
            val = config["constantParameters.parallelGain"]
            self._set("corrections.parallelGain.available", val)
            self._set("corrections.parallelGain.available", val)
            self._update_correction_flags()
        if config.has("corrections.parallelGain.sigmaHighMedium"):
            self._gain_sigma_high_medium = self._xp.float32(
                config["corrections.parallelGain.sigmaHighMedium"]
            )
        if config.has("corrections.parallelGain.sigmaMediumLow"):
            self._gain_sigma_medium_low = self._xp.float32(
                config["corrections.parallelGain.sigmaMediumLow"]
            )
        if config.has("corrections.parallelGain.handling"):
            self._parallel_gain_handling = ParallelGainOptions[
                config["corrections.parallelGain.handling"]
            ]

    def flush_buffers(self, constants):
        if Constants.Offset in constants:
            self.offset_map.fill(0)
        if Constants.GainAmpMap in constants:
            self.gain_amp_map.fill(1)
        if Constants.Noise in constants:
            self.noise_map.fill(0)
        if Constants.RelativeGain in constants:
            self.rel_gain_slopes_map.fill(1)
        if Constants.FFMap in constants:
            self.flatfield_map.fill(1)
        if constants & bad_pixel_constants:
            self.bad_pixel_map.fill(0)


class LpdGpuRunner(LpdBaseRunner):
    runner_type = base_kernel_runner.KernelRunnerTypes.GPU

    def _post_init(self):
        cuda_module = self._xp.RawModule(
            code=base_kernel_runner.get_kernel_template("lpd_gpu.cu").render(
                {
                    "ss_dim": self.num_pixels_ss,
                    "fs_dim": self.num_pixels_fs,
                    "output_data_dtype": utils.np_dtype_to_c_type(
                        self._output_dtype
                    ),
                    "corr_enum": utils.enum_to_c_template(self._correction_flag_class),
                }
            )
        )
        self.correction_kernel = cuda_module.get_function("correct")
        self.gain_threshold_kernel = cuda_module.get_function("parallel_gain_threshold")

    def _correct(self, flags, image_data, cell_table, processed_data, gain_map):
        num_frames = self._xp.uint16(image_data.shape[0])
        block_shape = (1, 1, 64)
        grid_shape = utils.grid_to_cover_shape_with_blocks(
            processed_data.shape, block_shape
        )

        if flags & CorrectionFlags.PARALLELGAIN:
            # note: processed_data has num_frames // 3 frames already
            # plan: override image_data with thresholded / sliced data
            num_frames = num_frames // 3

            if self._parallel_gain_handling is ParallelGainOptions.THRESHOLD:
                self.gain_threshold_kernel(
                    grid_shape,
                    block_shape,
                    (
                        num_frames,
                        cell_table,
                        self._constant_memory_cells,
                        self._gain_sigma_high_medium,
                        self._gain_sigma_medium_low,
                        self.offset_map,
                        self.noise_map,
                        image_data,
                    )
                )
                # kernel put results into first third of "frames" in buffer
                image_data = image_data[:num_frames]

            else:
                gain_stage = self._parallel_gain_handling.value
                image_data = image_data[
                    gain_stage * num_frames:(gain_stage + 1) * num_frames
                ]
            # ordinary kernel will later extract gain values,
            # so just leave them in "raw" data passed to it

        # regular kernel does not need to know about parallel gain
        self.correction_kernel(
            grid_shape,
            block_shape,
            (
                cell_table,
                flags,
                num_frames,
                self._constant_memory_cells,
                self.offset_map,
                self.gain_amp_map,
                self.rel_gain_slopes_map,
                self.flatfield_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                image_data,
                gain_map,
                processed_data,
            ),
        )


class LpdCpuRunner(LpdBaseRunner):
    runner_type = base_kernel_runner.KernelRunnerTypes.CPU

    def _post_init(self):
        from ..kernels import lpd_cython
        self.correction_kernel = lpd_cython.correct

    def _correct(self, flags, image_data, cell_table, processed_data, gain_map):
        self.correction_kernel(
            image_data,
            cell_table,
            flags,
            self.offset_map,
            self.gain_amp_map,
            self.rel_gain_slopes_map,
            self.flatfield_map,
            self.bad_pixel_map,
            self.bad_pixel_mask_value,
            gain_map,
            processed_data,
        )


class LpdCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {
            Constants.Offset: self.dark_condition,
            Constants.Noise: self.dark_condition,
            Constants.BadPixelsDark: self.dark_condition,
            Constants.GainAmpMap: self.gain_condition,
            Constants.FFMap: self.gain_condition,
            Constants.RelativeGain: self.gain_condition,
            Constants.BadPixelsFF: self.gain_condition,
        }

    @staticmethod
    def add_schema(schema):
        super(LpdCalcatFriend, LpdCalcatFriend).add_schema(schema, "LPD-Type")

        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(250)
            .commit(),
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.feedbackCapacitor")
            .tags("managed")
            .unit(Unit.FARAD)
            .metricPrefix(MetricPrefix.PICO)  # according to CalCat
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.photonEnergy")
            .tags("managed")
            .unit(Unit.ELECTRONVOLT)
            .metricPrefix(MetricPrefix.KILO)
            .assignmentOptional()
            .defaultValue(9.3)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.category")
            .tags("managed")
            .displayedName("Category")
            .unit(Unit.NUMBER)
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("constantParameters.useMemoryCellOrder")
            .tags("managed")
            .displayedName("Use memory cell order parameter")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.memoryCellOrder")
            .tags("managed")
            .displayedName("Memory cell order")
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("constantParameters.parallelGain")
            .tags("managed")
            .displayedName("Parallel gain mode")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def basic_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Feedback capacitor"] = self._get_param("feedbackCapacitor")

        return res

    def dark_condition(self):
        res = self.basic_condition()
        if self._get_param("useMemoryCellOrder"):
            res["Memory cell order"] = self._get_param("memoryCellOrder")
        if self._get_param("parallelGain"):
            res["Parallel gain"] = 1

        return res

    def gain_condition(self):
        res = self.basic_condition()
        res["Source Energy"] = self._get_param("photonEnergy")
        res["category"] = self._get_param("category")
        return res


@KARABO_CLASSINFO("LpdCorrection", deviceVersion)
class LpdCorrection(base_correction.BaseCorrection):
    _base_output_schema = schemas.xtdf_output_schema
    _correction_steps = correction_steps
    _kernel_runner_class = LpdGpuRunner
    _calcat_friend_class = LpdCalcatFriend
    _constant_enum_class = Constants

    _preview_outputs = [PreviewSpec(name) for name in ("raw", "corrected", "gainMap")]

    def __init__(self, config):
        super().__init__(config)
        self._parallel_gain = config["constantParameters.parallelGain"]

    @classmethod
    def expectedParameters(cls, expected):
        LpdCalcatFriend.add_schema(expected)
        LpdBaseRunner.add_schema(expected)
        base_correction.add_addon_nodes(expected, cls)
        PreviewFriend.add_schema(expected, cls._preview_outputs)
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(
                cls._base_output_schema(use_shmem_handles=cls._use_shmem_handles)
            )
            .commit(),
        )

    def _get_data_from_hash(self, data_hash):
        """Will get image data, cell table, pulse table, and list of other arrays from
        the data hash. Assumes XTDF (image.data, image.cellId, image.pulseId, and no
        other data), non-XTDF subclass can override. Cell and pulse table can be
        None."""

        image_data = data_hash.get(self._image_data_path)
        cell_table = data_hash[self._cell_table_path].ravel()
        pulse_table = data_hash[self._pulse_table_path].ravel()
        num_frames = cell_table.size

        if self._parallel_gain:
            # TODO: warn if num_frames not multiple of 3
            actual_frames = num_frames // 3
            cell_table[actual_frames : 2 * actual_frames] = cell_table[:actual_frames]
            cell_table[2 * actual_frames:] = cell_table[:actual_frames]
            pulse_table[actual_frames : 2 * actual_frames] = pulse_table[:actual_frames]
            pulse_table[2 * actual_frames:] = pulse_table[:actual_frames]

        # DataAggregator typically tells us the wrong axis order
        if self.unsafe_get("workarounds.overrideInputAxisOrder"):
            # note: regular correction kernel needn't know about parallel gain
            expected_shape = self.kernel_runner.expected_input_shape(
                num_frames
            )
            if expected_shape != image_data.shape:
                image_data.shape = expected_shape
        return (
            num_frames,
            image_data,
            cell_table,
            pulse_table,
        )

    def _maybe_update_cell_and_pulse_tables(self, cell_table, pulse_table):
        if self._parallel_gain and cell_table is not None:
            actual_frames = cell_table.size // 3
            cell_table = cell_table[:actual_frames]
            pulse_table = pulse_table[:actual_frames]
        super()._maybe_update_cell_and_pulse_tables(cell_table, pulse_table)

    def postReconfigure(self):
        super().postReconfigure()
        self._parallel_gain = self.get("constantParameters.parallelGain")
