import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OVERWRITE_ELEMENT,
)

from .. import base_kernel_runner
from .._version import version as deviceVersion
from . import LpdCorrection


class LpdminiGpuRunner(LpdCorrection.LpdGpuRunner):
    runner_type = base_kernel_runner.KernelRunnerTypes.GPU
    num_pixels_ss = 32

    def load_constant(self, constant_type, constant_data):
        # constant type → transpose order
        constant_buffer_map = {
            LpdCorrection.Constants.Offset: self.offset_map,
            LpdCorrection.Constants.GainAmpMap: self.gain_amp_map,
            LpdCorrection.Constants.FFMap: self.flatfield_map,
            LpdCorrection.Constants.RelativeGain: self.rel_gain_slopes_map,
        }
        if constant_type in {
            LpdCorrection.Constants.BadPixelsDark,
            LpdCorrection.Constants.BadPixelsFF,
        }:
            self.bad_pixel_map |= self._xp.asarray(
                constant_data,
                dtype=np.uint32,
            )[: self.constant_memory_cells]
        else:
            constant_buffer_map[constant_type].set(
                constant_data.astype(np.float32)[: self.constant_memory_cells]
            )


class LpdminiCalcatFriend(LpdCorrection.LpdCalcatFriend):
    @staticmethod
    def add_schema(schema):
        super(LpdminiCalcatFriend, LpdminiCalcatFriend).add_schema(schema)
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDisplayedName("Bias voltage (odd)")
            .setNewDescription("Bias voltage apbplied to minis 1, 3, 5, and 7.")
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.biasVoltage2")
            .tags("managed")
            .displayedName("Bias voltage (even)")
            .description("Separate bias voltage used for minis 2, 4, 6, and 8.")
            .assignmentOptional()
            .defaultValue(300)
            .reconfigurable()
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(32)
            .commit(),
        )

    def basic_condition(self):
        res = super().basic_condition()
        if int(self.device.get("fastSources")[0][-1]) % 2 == 0:
            res["Sensor Bias Voltage"] = self._get_param("biasVoltage2")
        if "category" in res:
            del res["category"]
        return res


@KARABO_CLASSINFO("LpdminiCorrection", deviceVersion)
class LpdminiCorrection(LpdCorrection.LpdCorrection):
    _calcat_friend_class = LpdminiCalcatFriend
    _kernel_runner_class = LpdminiGpuRunner
