import numpy as np
import tokenize
from token import (
    AMPER,
    CIRCUMFLEX,
    DEDENT,
    INDENT,
    ISEOF,
    LPAR,
    NAME,
    NEWLINE,
    NL,
    RPAR,
    TILDE,
    VBAR,
)
from collections import deque


class BaseExpression:

    def __init__(self, expr):
        self.expr = expr

    @classmethod
    def parse(cls, string, variables={}):
        if not string.strip():
            return cls(None)

        stack = deque()
        result = []
        token_iter = tokenize.generate_tokens(iter([string]).__next__)
        try:
            token = next(token_iter)
        except tokenize.TokenError:
            raise SyntaxError("invalid syntax")

        while not ISEOF(token.type):
            tok_type = token.exact_type

            if tok_type == LPAR:
                stack.append(token)
            elif tok_type in cls.UNARY:
                stack.append(token)
            elif tok_type == RPAR:
                try:
                    head = stack.pop()
                    while head.exact_type != LPAR:
                        result.append(head)
                        head = stack.pop()
                except IndexError:
                    raise SyntaxError(f"unmatched ')', pos: {token.start[1]}")
            elif tok_type in cls.BINARY:
                prio = cls.PRIORITY[tok_type] | cls.UNARY.keys()
                while stack and stack[-1].exact_type in prio:
                    result.append(stack.pop())
                stack.append(token)
            elif tok_type in cls.BLANK:
                pass
            elif tok_type == NAME:
                if token.string not in variables:
                    raise NameError(
                        f"name '{token.string}' is not defined, "
                        f"pos: {token.start[1]}"
                    )
                result.append(token)
            else:
                raise SyntaxError(f"invalid syntax, pos: {token.start[1]}")

            try:
                token = next(token_iter)
            except tokenize.TokenError:
                raise SyntaxError(f"invalid syntax, pos: {token.end[1] + 1}")

        while stack:
            head = stack.pop()
            if head.exact_type == LPAR:
                raise SyntaxError("unexpected EOF while parsing")
            result.append(head)

        for token in result:
            tok_type = token.exact_type
            if tok_type in cls.BINARY:
                stack.pop()
            elif tok_type == NAME:
                stack.append(token)

        if len(stack) != 1:
            raise SyntaxError(f"invalid syntax, pos: {stack[-1].end[1] + 1}")

        return cls(result)

    def eval(self, variables, default):
        if self.expr is None:
            return default
        stack = deque()
        for token in self.expr:
            tok_type = token.exact_type
            if tok_type in self.UNARY:
                op = self.UNARY[tok_type]
                stack.append(op(stack.pop()))
            elif tok_type in self.BINARY:
                op = self.BINARY[tok_type]
                b = stack.pop()
                a = stack.pop()
                stack.append(op(a, b))
            else:
                stack.append(variables[token.string])

        return stack.pop()


class BinaryExpression(BaseExpression):
    BLANK = {
        NEWLINE,
        NL,
        INDENT,
        DEDENT,
    }
    UNARY = {TILDE: np.logical_not}
    BINARY = {AMPER: np.logical_and, CIRCUMFLEX: np.logical_xor, VBAR: np.logical_or}
    PRIORITY = {
        AMPER: {AMPER},
        CIRCUMFLEX: {AMPER, CIRCUMFLEX},
        VBAR: {AMPER, CIRCUMFLEX, VBAR},
    }
