import re

from karabo.bound import (
    BOOL_ELEMENT,
    NODE_ELEMENT,
    STRING_ELEMENT,
    TABLE_ELEMENT,
    VECTOR_STRING_ELEMENT,
    Hash,
    Schema,
    VectorHash,
)
import numpy as np

from . import redu_op


def reduction_classes_schema():
    schema = Schema()
    (
        STRING_ELEMENT(schema)
        .key("operation")
        .displayedName("Operation")
        .readOnly()
        .commit(),

        STRING_ELEMENT(schema)
        .key("argument")
        .displayedName("Key to reduce")
        .readOnly()
        .commit(),

        STRING_ELEMENT(schema)
        .key("selection")
        .displayedName("Selection Name")
        .readOnly()
        .commit(),

        STRING_ELEMENT(schema).key("result")
        .displayedName("Result Key")
        .readOnly()
        .commit(),
    )
    return schema


class FrameselectionFriend:
    @staticmethod
    def add_schema(schema):
        (
            NODE_ELEMENT(schema)
            .key("frameSelector")
            .displayedName("Frame selection")
            .commit(),

            BOOL_ELEMENT(schema)
            .key("frameSelector.enable")
            .displayedName("Enabled")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("frameSelector.arbiterSource")
            .displayedName("Arbiter source")
            .description(
                "Source name to pull the frame selection pattern from, must be part of "
                "matched sources."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("frameSelector.dataSourcePattern")
            .displayedName("Data source pattern")
            .description(
                "Source name pattern to apply frame selection to. Should match "
                "subset of matched sources."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            VECTOR_STRING_ELEMENT(schema)
            .key("frameSelector.dataKeys")
            .displayedName("Data keys")
            .description("Keys in data sources to apply frame selection to.")
            .assignmentOptional()
            .defaultValue([])
            .reconfigurable()
            .commit(),

            TABLE_ELEMENT(schema).key("frameSelector.reduction")
            .displayedName("Reduction")
            .description(
                "The reduction operations on selected images according to "
                "the input from Arbiter.")
            .setColumns(reduction_classes_schema())
            .readOnly().initialValue([])
            .commit(),
        )

    def __init__(self, device, config):
        self._device = device
        self._config = Hash()
        self._enabled = False
        self._arbiter = ""
        self._data_pattern = ""
        self._data_keys = []
        self._reduction_classes = {}
        self.reconfigure(config)
        self._device.KARABO_SLOT(self.receive_reduction_classes)

    def reconfigure(self, config):
        self._config.merge(config)
        self._enabled = self._config.get("enable")
        self._arbiter = self._config.get("arbiterSource")
        self._source_pattern = re.compile(self._config.get("dataSourcePattern"))
        self._data_keys = list(self._config.get("dataKeys"))

    def receive_reduction_classes(self, redu_classes):
        reduction = VectorHash()
        self._reduction_classes = {}
        for encoded_class in redu_classes:
            name, op, arg, result = encoded_class.split(':')
            reduction.append(Hash(
                "operation", op,
                "argument", arg,
                "selection", name,
                "result", result,
            ))
            self._reduction_classes[result] = (op, arg, name)
        self._device.set("frameSelector.reduction", reduction)

    def get_mask(self, sources):
        if self._enabled and self._arbiter in sources:
            return np.array(
                sources[self._arbiter][0]["data.dataFramePattern"], copy=False
            ).astype(bool, copy=False)
        else:
            return None

    def apply_mask(self, source, data_hash, mask):
        if mask is not None and self._source_pattern.match(source):
            for key in self._data_keys:
                if not data_hash.has(key):
                    continue
                if data_hash[key].shape[0] != mask.size:
                    self._device.log.WARN(
                        "Frame selection mask does not match the data size"
                    )
                    continue

                data_hash[key] = data_hash[key][mask]

    def resolve_reduction_classes(self, sources):
        correction_sources = {
            source: data
            for source, (data, timestamp) in sources.items()
            if self._source_pattern.match(source)
        }
        operations = []
        if self._enabled and self._arbiter in sources:
            arbiter = sources[self._arbiter][0]
            for result, (op, arg, selection) in self._reduction_classes.items():
                mask_key = f"{selection}.mask"
                if not arbiter.has(mask_key):
                    # no mask
                    continue
                mask = np.array(
                    arbiter[mask_key], dtype=bool, copy=False)

                for source, data in correction_sources.items():
                    if not data.has(arg):
                        # no array to reduce
                        continue
                    if data[arg].shape[0] != mask.size:
                        self._device.log.WARN(
                            f"The reduction mask '{selection}' does not match the '{arg}' size"
                        )
                        continue

                    op_class = getattr(redu_op, op, None)
                    if op_class is None:
                        # unknown operation
                        continue
                    operations.append(op_class(data, arg, mask, result))
        return operations
