template<typename T>
class NeumaierSum {
    // https://en.wikipedia.org/wiki/Kahan_summation_algorithm
private:
    T sum = 0;
    T c = 0;
public:
    __device__ void add(const T val) {
        const T t = sum + val;
        if (fabs(sum) >= fabs(val)) {
            c += (sum - t) + val;
        } else {
            c += (val - t) + sum;
        }
        sum = t;
    }

    __device__ T get() {
        return sum + c;
    }
};

template<typename image_t, typename csr_t>
__global__ void integrate_1d_csr(
		const int csr_nrows, // = #rings
        const int csr_ncols, // = #pixels
		const csr_t* csr_data, // fixed dtype for now
        const int* csr_indices,
        const int* csr_indptr,
        const int n_frames,
        const image_t* frames, // #frames x #pixels
        image_t* output  // #frames x #rings
    ) {
    // this thread handles one element in output matrix
    const int output_frame = blockDim.x * blockIdx.x + threadIdx.x;
    const int output_ring = blockDim.y * blockIdx.y + threadIdx.y;
    if (output_frame >= n_frames || output_ring >= csr_nrows) {
        return;
    }

    NeumaierSum<image_t> value;
    NeumaierSum<image_t> weight;

    const int frame_start = output_frame * csr_ncols;
    for (int i=csr_indptr[output_ring]; i<csr_indptr[output_ring + 1]; ++i) {
        // read data
        const image_t pixel_value = frames[frame_start + csr_indices[i]];
        const float pixel_weight = csr_data[i];

        if (isnan(pixel_value)) {
            continue;
        }

        value.add(pixel_weight * pixel_value);
        weight.add(pixel_weight);
    }
    const float final_weight = weight.get();
	output[output_frame * csr_nrows + output_ring] =
		final_weight > 0 ?
		value.get() / final_weight :
		nan("0");
}

/* This version does *not* divide by the weight included in a given ring. Instead, this
   value is output along with the un-normalized weighted pixel total s.t. a downstream
   device combining data from multiple modules can normalize properly i.e. sum all the
   mass in the ring and the weight and divide the former by the latter. */
template<typename image_t, typename csr_t>
__global__ void integrate_1d_csr_multimodule(
		const int csr_nrows, // = #rings (bins is this + 1)
        const int csr_ncols, // = #pixels per frame
		const float* csr_data, // fixed dtype for now
        const int* csr_indices,
		const int* csr_indptr,  // size csr_nrows + 1
        const int n_frames,
        const image_t* frames, // #frames x #pixels
		image_t* output_total,  // #frames x #rings
        image_t* output_weight  // #frames x #rings
    ) {
    // this thread handles one element in output matrix
    const int output_frame = blockDim.x * blockIdx.x + threadIdx.x;
    const int output_ring = blockDim.y * blockIdx.y + threadIdx.y;
    if (output_frame >= n_frames || output_ring >= csr_nrows) {
        return;
    }

    NeumaierSum<image_t> value;
    NeumaierSum<image_t> weight;

    const int frame_start = output_frame * csr_ncols;
    for (int i=csr_indptr[output_ring]; i<csr_indptr[output_ring + 1]; ++i) {
        // read data
        const image_t pixel_value = frames[frame_start + csr_indices[i]];
        const float pixel_weight = csr_data[i];

        if (isnan(pixel_value)) {
            continue;
        }

        value.add(pixel_weight * pixel_value);
        weight.add(pixel_weight);
    }
    const float final_weight = weight.get();
    output_total[output_frame * csr_nrows + output_ring] = final_weight > 0 ? value.get() : nan("0");
    output_weight[output_frame * csr_nrows + output_ring] = final_weight > 0 ? final_weight : nan("0");
}
