# distutils: language=c++
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: nonecheck=False
cimport cython
cimport numpy as np
cimport openmp
from cython.parallel cimport prange
from libc.math cimport NAN, atan2, isfinite, sqrt
from libcpp.algorithm cimport upper_bound
from libcpp.iterator cimport distance
from libcpp.vector cimport vector

import numpy as np
import scipy.sparse


# TODO: get rid of this guy in favor of bare memoryviews
cdef struct vec3_t:
    double x
    double y
    double z


cdef vec3_t scale_arr3(double[:] v, double s) noexcept nogil:
    return vec3_t(s * v[0], s * v[1], s * v[2])


cdef vec3_t lin_comb_vec3(vec3_t v, double s1, vec3_t v1, double s2, vec3_t v2) noexcept nogil:
    return vec3_t(
        v.x + s1 * v1.x + s2 * v2.x,
        v.y + s1 * v1.y + s2 * v2.y,
        v.z + s1 * v1.z + s2 * v2.z,
    )


cdef vec3_t lin_comb_arr_arr(double[:] v, double s1, double[:] v1, double s2, double[:] v2) noexcept nogil:
    return vec3_t(
        v[0] + s1 * v1[0] + s2 * v2[0],
        v[1] + s1 * v1[1] + s2 * v2[1],
        v[2] + s1 * v1[2] + s2 * v2[2],
    )


cdef double point_theta(vec3_t v) noexcept nogil:
    return atan2(
        sqrt(v.x * v.x + v.y * v.y),
        v.z,
    )


ctypedef fused bin_t:
    float
    double


cdef struct BinAndWeight:
    Py_ssize_t bin_index
    float weight


cdef vector[BinAndWeight] bin_pixel_subsample(
    bin_t[:] bins,
    vec3_t pixel_corner,
    vec3_t sub_step_ss,
    vec3_t sub_step_fs,
    Py_ssize_t n_split,
) noexcept nogil:
    cdef Py_ssize_t sub_index_ss, sub_index_fs  # indices for subsample steps
    cdef vec3_t sample_point  # subpixel point we're placing in a bin
    cdef float sample_point_theta
    cdef Py_ssize_t bin_index_found
    # taking half a step into pixel so we don't bias towards "lower left" corner
    cdef vec3_t start_point = lin_comb_vec3(pixel_corner, 0.5, sub_step_ss, 0.5, sub_step_fs)
    cdef vector[Py_ssize_t] bin_counts  # "custom bin counter"
    cdef vector[BinAndWeight] res
    # given n bins, output "bin indices" will be from 0 to n-2
    # a sample with theta contributes to output "bin index" i if b[i] <= theta b[i+1]
    # hence the upper_bound and - 1 later
    # sample points outside range are ignored
    bin_counts.resize(bins.shape[0] - 1)  # full of zeros initially
    # and more loops in order to subsample...
    for sub_index_ss in range(n_split):
        for sub_index_fs in range(n_split):
            sample_point = lin_comb_vec3(
                start_point,
                <double>sub_index_ss, sub_step_ss,
                <double>sub_index_fs, sub_step_fs
            )
            sample_point_theta = point_theta(sample_point)
            if sample_point_theta < bins[0] or sample_point_theta >= bins[bins.shape[0]-1]:
                continue
            bin_index_found = distance(
                &bins[0], upper_bound(&bins[0], &bins[0] + bins.shape[0], sample_point_theta)
            )
            # note: with previous check distance to upper_bound should be in [1, #bins-2]
            bin_counts[bin_index_found - 1] += 1
    cdef Py_ssize_t elem = 0
    for count in bin_counts:
        if count > 0:
            res.push_back(BinAndWeight(elem, <float>count / <float>(n_split * n_split)))
        elem += 1
    return res


def generate_coo(
    bin_t[:] bins,
    double[:, ::1] corners,
    double[:, ::1] ss_vecs,
    double[:, ::1] fs_vecs,
    Py_ssize_t pixels_per_tile_ss,
    Py_ssize_t pixels_per_tile_fs,
    Py_ssize_t n_split,
):
    cdef Py_ssize_t num_pixels = corners.shape[0] * pixels_per_tile_ss * pixels_per_tile_fs
    cdef vector[Py_ssize_t] row_vec, col_vec
    cdef vector[float] data_vec
    cdef openmp.omp_lock_t lock
    openmp.omp_init_lock(&lock)
    # could maybe come up with better estimate, but having to allocate more later is not
    # the end of the world
    row_vec.reserve(num_pixels * 2)
    col_vec.reserve(num_pixels * 2)
    data_vec.reserve(num_pixels * 2)

    cdef Py_ssize_t tile_index, pixel_ss, pixel_fs, pixel_index
    cdef vec3_t sub_step_ss, sub_step_fs # (ss_vec and fs_vec) / n_split
    # these two just for iterating over bin_pixel_subsample results
    cdef BinAndWeight bin_and_weight
    cdef vector[BinAndWeight] subsample_res

    for tile_index in prange(corners.shape[0], nogil=True):
        sub_step_ss = scale_arr3(ss_vecs[tile_index], 1.0 / <float>n_split)
        sub_step_fs = scale_arr3(fs_vecs[tile_index], 1.0 / <float>n_split)
        for pixel_ss in prange(pixels_per_tile_ss):
            for pixel_fs in range(pixels_per_tile_fs):
                pixel_index = tile_index * (pixels_per_tile_ss * pixels_per_tile_fs) + pixel_ss * pixels_per_tile_fs + pixel_fs
                subsample_res = bin_pixel_subsample(
                    bins,
                    lin_comb_arr_arr(
                        corners[tile_index],
                        <float>pixel_ss, ss_vecs[tile_index],
                        <float>pixel_fs, fs_vecs[tile_index],
                    ), # corner of this pixel
                    sub_step_ss,
                    sub_step_fs,
                    n_split,
                )
                openmp.omp_set_lock(&lock)
                for bin_and_weight in subsample_res:
                    row_vec.push_back(bin_and_weight.bin_index)
                    col_vec.push_back(pixel_index)
                    data_vec.push_back(bin_and_weight.weight)
                openmp.omp_unset_lock(&lock)
    # TODO: contemplate ownership; is this safe and copy-free?
    return np.asarray(data_vec), np.asarray(row_vec), np.asarray(col_vec)


def integrate_1d_csr(
    int[:] csr_indices,
    int[:] csr_indptr,
    float[:] csr_data,
    float[:, ::1] frames,
    float[:, ::1] output,
):
    cdef Py_ssize_t frame_index, ring_index, data_index
    cdef float pixel_acc, weight_acc, weight_val, pixel_val

    for frame_index in range(frames.shape[0]):
        for ring_index in range(csr_indptr.shape[0] - 1):
            weight_acc = 0
            pixel_acc = 0
            for data_index in range(csr_indptr[ring_index], csr_indptr[ring_index+1]):
                pixel_val = frames[frame_index, csr_indices[data_index]]
                if not isfinite(pixel_val):
                    continue
                weight_val = csr_data[data_index]
                weight_acc = weight_acc + weight_val
                pixel_acc = pixel_acc + weight_val * pixel_val
            output[frame_index, ring_index] = pixel_acc / weight_acc if weight_acc > 0 else NAN


def integrate_1d_csr_multimodule(
    int[:] csr_indices,
    int[:] csr_indptr,
    float[:] csr_data,
    float[:, ::1] frames,
    float[:, ::1] output_total,
    float[:, ::1] output_weight,
):
    cdef Py_ssize_t frame_index, ring_index, data_index
    cdef float pixel_acc, weight_acc, weight_val, pixel_val

    for frame_index in range(frames.shape[0]):
        for ring_index in range(csr_indptr.shape[0] - 1):
            weight_acc = 0
            pixel_acc = 0
            for data_index in range(csr_indptr[ring_index], csr_indptr[ring_index+1]):
                pixel_val = frames[frame_index, csr_indices[data_index]]
                if not isfinite(pixel_val):
                    continue
                weight_val = csr_data[data_index]
                weight_acc = weight_acc + weight_val
                pixel_acc = pixel_acc + weight_val * pixel_val
            output_total[frame_index, ring_index] = pixel_acc if weight_acc > 0 else NAN
            output_weight[frame_index, ring_index] = weight_acc if weight_acc > 0 else NAN
