#include <cuda_fp16.h>

{{corr_enum}}

extern "C" {
	__global__ void correct(const unsigned short* cell_table,
	                        const unsigned char corr_flags,
	                        const unsigned short input_frames,
	                        const unsigned short map_memory_cells,
	                        const float* offset_map, // shape: cell, ss, fs, gain
	                        const float* gain_amp_map,
	                        const float* rel_gain_slopes_map,
	                        const float* flatfield_map,
	                        const unsigned int* bad_pixel_map,
	                        const float bad_pixel_mask_value,
	                        // note: will overwrite data with original value stripped of gain bits
	                        unsigned short* data, // shape: memory cell, 1, ss, fs
	                        float* gain_map, // similar to preview for AGIPD
	                        {{output_data_dtype}}* output) {
		const size_t fs_dim = {{fs_dim}};
		const size_t ss_dim = {{ss_dim}};

		const size_t frame = blockIdx.x * blockDim.x + threadIdx.x;
		const size_t ss = blockIdx.y * blockDim.y + threadIdx.y;
		const size_t fs = blockIdx.z * blockDim.z + threadIdx.z;

		if (frame >= input_frames || ss >= ss_dim || fs >= fs_dim) {
			return;
		}

		const size_t data_stride_fs = 1;
		const size_t data_stride_ss = fs_dim * data_stride_fs;
		const size_t data_stride_frame = ss_dim * data_stride_ss;
		const size_t data_index = frame * data_stride_frame + ss * data_stride_ss + fs * data_stride_fs;
		const unsigned short raw_data_value = data[data_index];
		const unsigned char gain = (raw_data_value >> 12) & 0x0003;
		const unsigned short raw_data_no_gain = raw_data_value & 0x0fff;
		float corrected = (float)(raw_data_no_gain);
		float gain_for_preview = (float)gain;

		const size_t gm_map_stride_gain = 1;
		const size_t gm_map_stride_fs = 3 * gm_map_stride_gain;
		const size_t gm_map_stride_ss = fs_dim * gm_map_stride_fs;
		const size_t gm_map_stride_cell = ss_dim * gm_map_stride_ss;

		const size_t map_cell = cell_table[frame];
		if (map_cell < map_memory_cells) {
			const size_t gm_map_index = gain * gm_map_stride_gain +
				map_cell * gm_map_stride_cell +
				ss * gm_map_stride_ss +
				fs * gm_map_stride_fs;

			if (gain > 2 || ((corr_flags & BPMASK) && bad_pixel_map[gm_map_index])) {
				// now also checking for illegal gain value
				corrected = bad_pixel_mask_value;
				gain_for_preview = bad_pixel_mask_value;
			} else {
				// TODO: save GPU memory by combining maps
				if (corr_flags & OFFSET) {
					corrected -= offset_map[gm_map_index];
				}
				if (corr_flags & GAIN_AMP) {
					corrected *= gain_amp_map[gm_map_index];
				}
				if (corr_flags & REL_GAIN) {
					corrected *= rel_gain_slopes_map[gm_map_index];
				}
				if (corr_flags & FF_CORR) {
					corrected *= flatfield_map[gm_map_index];
				}
				if (corrected < -1e7 || corrected > 1e7 || isnan(corrected) || isinf(corrected)) {
					corrected = bad_pixel_mask_value;
				}
			}
		} else {
			if (gain > 2 || corrected < -1e7 || corrected > 1e7) {
				corrected = bad_pixel_mask_value;
				gain_for_preview = bad_pixel_mask_value;
			}
		}

		gain_map[data_index] = gain_for_preview;
		data[data_index] = raw_data_no_gain;
		{% if output_data_dtype == "half" %}
		output[data_index] = __float2half(corrected);
		{% else %}
		output[data_index] = ({{output_data_dtype}})corrected;
		{% endif %}
	}

	__global__ void parallel_gain_threshold(const unsigned short output_frames,  // one third of input frames
	                                        const unsigned short* cell_table,
	                                        const unsigned short map_memory_cells,
	                                        const float threshold_sigma_high_med,
	                                        const float threshold_sigma_med_low,
											const float* offset_map, // shape: cell, ss, fs, gain
											const float* noise_map, // shape: cell, ss, fs, gain
	                                        // data is input *and* output
	                                        // shape: memory cell, 1, ss, fs
	                                        unsigned short* data) {
		const size_t fs_dim = {{fs_dim}};
		const size_t ss_dim = {{ss_dim}};

		const size_t frame = blockIdx.x * blockDim.x + threadIdx.x;
		const size_t ss = blockIdx.y * blockDim.y + threadIdx.y;
		const size_t fs = blockIdx.z * blockDim.z + threadIdx.z;

		if (frame >= output_frames || ss >= ss_dim || fs >= fs_dim) {
			return;
		}

		const size_t map_cell = cell_table[frame];
		if (map_cell >= map_memory_cells) {
			return;
		}

		const size_t data_stride_fs = 1;
		const size_t data_stride_ss = fs_dim * data_stride_fs;
		const size_t data_stride_frame = ss_dim * data_stride_ss;
		const size_t data_stride_gain = output_frames * data_stride_frame;

		const size_t data_index_hg = frame * data_stride_frame + ss * data_stride_ss + fs * data_stride_fs;
		const size_t data_index_mg = data_index_hg + data_stride_gain;
		const size_t data_index_lg = data_index_mg + data_stride_gain;

		const size_t gm_map_stride_gain = 1;
		const size_t gm_map_stride_fs = 3 * gm_map_stride_gain;
		const size_t gm_map_stride_ss = fs_dim * gm_map_stride_fs;
		const size_t gm_map_stride_cell = ss_dim * gm_map_stride_ss;

		const size_t gm_map_index_hg = map_cell * gm_map_stride_cell + ss * gm_map_stride_ss + fs * gm_map_stride_fs;
		const size_t gm_map_index_mg = gm_map_index_hg + gm_map_stride_gain;
		//const size_t gm_map_index_lg = gm_map_index_mg + gm_map_stride_gain; not used

		unsigned short data_value = data[data_index_hg];
		// stripping gain bit(s) for comparisons
		float data_value_float = static_cast<float>(data_value & 0x0fff);

		// compared to notebook: this is before offset is subtracted
		const float threshold_high_med = 4096 - (threshold_sigma_high_med * noise_map[gm_map_index_hg]);
		// compared to notebook: without gain bit, we are 4096 lower
		const float threshold_med_low = 4096 - (threshold_sigma_med_low * noise_map[gm_map_index_mg]);

		// TODO: handle NaNs
		if (data_value_float > threshold_high_med) {
			data_value = data[data_index_mg];
			data_value_float = static_cast<float>(data_value & 0x0fff);

			if (data_value_float > threshold_med_low) {
				data_value = data[data_index_lg];
			}
		}
		data[data_index_hg] = data_value;
	}
}
