import contextlib
import pathlib
import threading

import h5py
import numpy as np
from calngUtils import device as device_utils
from calngUtils.misc import ChainHash
from karabo.bound import Hash, Schema


def maybe_get(b):
    if hasattr(b, "get"):
        return b.get()
    else:
        return b


class DummyLogger:
    DEBUG = print
    INFO = print
    WARN = print


@device_utils.with_config_overlay
@device_utils.with_unsafe_get
class DummyCorrectionDevice:
    log = DummyLogger()

    def log_status_info(self, msg):
        self.log.INFO(msg)

    def log_status_warn(self, msg):
        self.log.WARN(msg)

    def get(self, key):
        return self._parameters.get(key)

    def set(self, key, value):
        self._parameters[key] = value

    def reconfigure(self, config):
        with self.new_config_context(config):
            self.runner.reconfigure(config)
        self._parameters.merge(config)

    def __init__(self, *friends):
        self._parameters = Hash()
        s = Schema()
        self._temporary_config = []
        self._temporary_config_lock = threading.RLock()
        for friend in friends:
            if isinstance(friend, dict):
                friend_class = friend.pop("friend_class")
                extra_args = friend
            else:
                friend_class = friend
                extra_args = {}
            friend_class.add_schema(s, **extra_args)
        h = s.getParameterHash()
        for path in h.getPaths():
            if h.hasAttribute(path, "defaultValue"):
                if "output" in path.split("."):
                    continue
                self._parameters[path] = h.getAttribute(path, "defaultValue")

        self.warnings = {}

    def _please_send_me_cached_constants(self, constants, callback):
        pass

    @contextlib.contextmanager
    def warning_context(self, key, warn_type):
        def aux(message):
            print(message)
            self.warnings[(key, warn_type)] = message

        yield aux

    @contextlib.contextmanager
    def new_config_context(self, config):
        with self._temporary_config_lock:
            old_config = self._parameters
            if isinstance(self._parameters, ChainHash):
                # in case we need to support recursive case / configuration stacking
                self._parameters = ChainHash(config, *old_config._hashes)
            else:
                self._parameters = ChainHash(config, old_config)
            yield
            self._parameters = old_config


caldb_store = pathlib.Path("/gpfs/exfel/d/cal/caldb_store")


def get_constant_from_file(path_to_file, file_name, data_set_name):
    """Intended to work with copy-pasting paths from CalCat. Specifially, the "Path to
    File", "File Name", and "Data Set Name" fields displayed for a given CCV."""
    with h5py.File(caldb_store / path_to_file / file_name, "r") as fd:
        return np.array(fd[f"{data_set_name}/data"])


class OverrideBase:
    overrides = tuple()

    @classmethod
    @contextlib.contextmanager
    def override_context(cls, other_cls):
        cache = {}
        for meth_name in cls.overrides:
            cache[meth_name] = getattr(other_cls, meth_name)
            setattr(other_cls, meth_name, getattr(cls, meth_name))
        try:
            yield
        finally:
            for meth_name in cls.overrides:
                setattr(other_cls, meth_name, cache[meth_name])


class OverridePythonDevice(OverrideBase):
    overrides = ("writeChannel",)

    def writeChannel(self, output_name, out_hash, timestamp=None):
        self._last_written_channel = output_name
        self._last_written_data = out_hash
