import extra_geom
import itertools
import numpy as np

from calng import azim_utils
from calng.utils import maybe_get


det_dist = 0.23
num_bins = 1000
num_frames = 352
geom = extra_geom.AGIPD_1MGeometry.example()
bins = np.linspace(0, azim_utils.get_geom_max_theta(geom, det_dist), 1000)
bin_midpoints = (bins[1:] + bins[:-1]) / 2


def angle_to_intensity(angle, dtype):
    return (
        np.exp(-((angle.min() - angle * 10) ** 2))
        + np.exp(-np.abs(np.quantile(angle, 0.6) - angle) ** 1.5 * 1000) * 10
    ).astype(dtype, copy=False)


def generate_frame(geom, dtype):
    ppos = geom.get_pixel_positions().reshape(-1, 3).astype(dtype, copy=False)
    ppos[..., 2] += det_dist
    angle = np.arctan2(
        np.linalg.norm(ppos[..., :2], axis=-1),
        ppos[..., 2],
    )
    return angle_to_intensity(angle, dtype)


for ai_class, scalar_dtype in itertools.product(
    (azim_utils.CpuAzimuthalIntegrator,),  # azim_utils.GpuAzimuthalIntegrator),
    (np.float32,),  # np.float64),
):
    print(f"Test round: {ai_class.__name__} with {scalar_dtype} data")
    frame = generate_frame(geom, scalar_dtype)

    ai = ai_class(geom, bins, det_dist, 1)

    frame = ai._xp.asarray(frame)

    # input with no frame axis -> output with no frame axis
    output = ai.integrate_1d(frame)

    assert output.shape == (bins.size - 1,)
    assert output.dtype == frame.dtype

    output = ai.integrate_1d(frame.reshape(geom.expected_data_shape))

    assert output.shape == (bins.size - 1,)
    assert output.dtype == frame.dtype

    # input with frame axis -> output with frame axis
    output = ai.integrate_1d(frame.reshape(1, frame.size))

    assert output.shape == (
        1,
        bins.size - 1,
    )
    assert output.dtype == frame.dtype

    output = ai.integrate_1d(frame.reshape(1, *geom.expected_data_shape))

    assert output.shape == (
        1,
        bins.size - 1,
    )
    assert output.dtype == frame.dtype

    del frame

    frames = ai._xp.random.random_sample(
        (num_frames,) + geom.expected_data_shape
    ).astype(scalar_dtype, copy=False)

    # input with frame axis -> output with frame axis
    output = ai.integrate_1d(frames.reshape(num_frames, -1))

    assert output.shape == (
        num_frames,
        bins.size - 1,
    )
    assert output.dtype == frames.dtype

    output = ai.integrate_1d(frames.reshape(num_frames, *geom.expected_data_shape))

    assert output.shape == (
        num_frames,
        bins.size - 1,
    )
    assert output.dtype == frames.dtype
