import numpy as np
import pytest
from calng.corrections.DsscCorrection import (
    Constants,
    DsscBaseRunner,
    DsscCalcatFriend,
    DsscCpuRunner,
    DsscGpuRunner,
)
from karabo.bound import Hash

from common_setup import DummyCorrectionDevice, maybe_get

output_dtype = np.float32
corr_dtype = np.float32
pixels_x = 512
pixels_y = 128
memory_cells = 400
num_frames = memory_cells
offset_map = (
    np.random.random(size=(pixels_x, pixels_y, memory_cells)).astype(corr_dtype) * 20
)
cell_table = np.arange(memory_cells, dtype=np.uint16)
np.random.shuffle(cell_table)
raw_data = np.random.randint(
    low=0, high=2000, size=(memory_cells, 1, pixels_y, pixels_x), dtype=np.uint16
)


# TODO: gather CPU implementations elsewhere
def correct_naive(data, cell_table, offset_map):
    corr = np.squeeze(data).astype(corr_dtype, copy=True)
    safe_cell_bool = cell_table < offset_map.shape[-1]
    safe_cell_index = cell_table[safe_cell_bool]
    corr[safe_cell_bool] -= offset_map.transpose()[safe_cell_index]
    return corr.astype(output_dtype, copy=False)


corrected_data = correct_naive(raw_data, cell_table, offset_map)
only_cast_data = np.squeeze(raw_data).astype(output_dtype)


@pytest.fixture(params=[DsscCpuRunner, DsscGpuRunner])
def kernel_runner(request):
    device = DummyCorrectionDevice(DsscBaseRunner, DsscCalcatFriend)
    runner = request.param(device)
    device.runner = runner
    return runner


def test_only_cast(kernel_runner):
    assert raw_data.shape == kernel_runner.expected_input_shape(num_frames)
    # note: previews are just buffers, not yet reduced
    _, processed_data, (preview_raw, preview_corrected) = kernel_runner.correct(
        raw_data, cell_table
    )
    assert np.allclose(maybe_get(processed_data), only_cast_data)
    assert np.allclose(maybe_get(preview_raw), only_cast_data)
    assert np.allclose(maybe_get(preview_corrected), only_cast_data)


def test_correct(kernel_runner):
    for preview in ("raw", "corrected"):
        kernel_runner._device.reconfigure(
            Hash(
                f"preview.{preview}.index",
                0,
                f"preview.{preview}.selectionMode",
                "frame",
            )
        )
    kernel_runner.load_constant(Constants.Offset, offset_map)
    _, processed_data, (preview_raw, preview_corrected) = kernel_runner.correct(
        raw_data, cell_table
    )
    assert np.allclose(maybe_get(processed_data), corrected_data)
    assert np.allclose(maybe_get(preview_raw), raw_data[:, 0])
    assert np.allclose(maybe_get(preview_corrected), corrected_data)

    kernel_runner._device.reconfigure(Hash("corrections.offset.enable", False))
    _, processed_data, (preview_raw, preview_corrected) = kernel_runner.correct(
        raw_data, cell_table
    )
    assert np.allclose(maybe_get(processed_data), np.squeeze(raw_data))
    assert np.allclose(maybe_get(preview_raw), np.squeeze(raw_data))
    assert np.allclose(maybe_get(preview_corrected), corrected_data)

    kernel_runner._device.reconfigure(
        Hash(
            "corrections.offset.enable",
            True,
            "corrections.offset.preview",
            False,
        )
    )
    _, processed_data, (preview_raw, preview_corrected) = kernel_runner.correct(
        raw_data, cell_table
    )
    assert np.allclose(maybe_get(processed_data), corrected_data)
    assert np.allclose(maybe_get(preview_raw), np.squeeze(raw_data))
    assert np.allclose(maybe_get(preview_corrected), np.squeeze(raw_data))


def test_correct_oob_cells(kernel_runner):
    wild_cell_table = cell_table * 2
    reference = correct_naive(raw_data, wild_cell_table, offset_map)
    kernel_runner.load_constant(Constants.Offset, offset_map)
    _, processed_data, previews = kernel_runner.correct(
        raw_data, wild_cell_table
    )
    assert np.allclose(maybe_get(processed_data), reference)
