import cupy as cp
import numpy as np
import pytest
from calng.corrections.LpdCorrection import (
    Constants,
    LpdBaseRunner,
    LpdCalcatFriend,
    LpdCpuRunner,
    LpdGpuRunner,
)
from karabo.bound import Hash

from common_setup import DummyCorrectionDevice


def assert_allclose(a, b):
    if isinstance(a, cp.ndarray):
        if isinstance(b, cp.ndarray):
            return cp.testing.assert_allclose(a, b)
        else:
            a = a.get()
            return np.testing.assert_allclose(a, b)
    if isinstance(b, cp.ndarray):
        b = b.get()
    return np.testing.assert_allclose(a, b)


@pytest.fixture(params=[LpdCpuRunner, LpdGpuRunner])
def kernel_runner(request):
    device = DummyCorrectionDevice(LpdBaseRunner, LpdCalcatFriend)
    return request.param(device)


@pytest.fixture
def cpu_runner():
    device = DummyCorrectionDevice(LpdBaseRunner, LpdCalcatFriend)
    return LpdCpuRunner(device)


@pytest.fixture
def gpu_runner():
    device = DummyCorrectionDevice(LpdBaseRunner, LpdCalcatFriend)
    return LpdGpuRunner(device)


class TestParallelGain:
    @classmethod
    def setup_class(cls):
        nframes = 12
        cells = cp.arange(nframes * 3, dtype=np.uint16)
        cells[nframes:] = 1234  # only first third / repetition is reliable
        cls.cells = cells
        shape = (nframes, 256, 256)
        base_data = (
            (cp.arange(np.prod(shape)) * 1000).astype(cp.uint16).reshape(shape)
        )
        cls.raw_image = cp.concatenate(
            [
                base_data,  # HG
                (base_data + cp.random.random(base_data.shape) * 20).astype(
                    cp.int16
                ),  # MG
                (base_data + cp.random.random(base_data.shape) * 20).astype(
                    cp.int16
                ),  # LG
            ]
        ).astype(np.uint16)
        cls.raw_image %= 2**12
        gain_data = np.zeros_like(cls.raw_image)
        gain_data[nframes : nframes * 2] = 1
        gain_data[nframes * 2 :] = 2
        cls.raw_data = cls.raw_image.copy()
        cls.raw_data |= gain_data

    def test_threshold(self, gpu_runner):
        t1 = 1000
        t2 = 2001
        gpu_runner.reconfigure(
            Hash(
                "constantParameters.parallelGain",
                True,
                "corrections.parallelGain.thresholdHighMedium",
                t1,
                "corrections.parallelGain.thresholdMediumLow",
                t2,
            )
        )
        corrections, processed, (_, preview_corrected, preview_gain) = (
            gpu_runner.correct(self.raw_data, self.cells)
        )
        nframes = self.cells.size // 3

        # now we do thresholding ourselves
        # start by assuming high gain
        tmp = self.raw_image.astype(np.float32)
        gain_map = cp.zeros((nframes,) + self.raw_image.shape[1:], dtype=np.uint8)
        # then bump to medium gain if above threshold
        gain_map[tmp[:nframes] > np.float32(t1)] = 1
        # now, some pixels are high and some medium
        # gain_map[(gain_map == 0) & (tmp[:nframes] > t2)] = 2
        # finally, bump pixels from medium to low gain if above t2
        gain_map[(gain_map == 1) & (tmp[nframes : nframes * 2] > np.float32(t2))] = 2

        # creating masks for making assertions
        high_gain_mask = gain_map == 0
        med_gain_mask = gain_map == 1
        low_gain_mask = gain_map == 2

        # sanity check: some pixels should go in each gain stage
        assert high_gain_mask.sum() > 0
        assert med_gain_mask.sum() > 0
        assert low_gain_mask.sum() > 0

        assert_allclose(preview_gain[high_gain_mask], 0)
        assert_allclose(preview_gain[med_gain_mask], 1)
        assert_allclose(preview_gain[low_gain_mask], 2)

        assert_allclose(processed[high_gain_mask], tmp[:nframes][high_gain_mask])
        assert_allclose(processed[high_gain_mask], tmp[:nframes][high_gain_mask])
        assert_allclose(processed[high_gain_mask], tmp[:nframes][high_gain_mask])


class TestAdaptiveGain:
    @classmethod
    def setup_class(cls):
        # generate some "raw data"
        raw_data_shape = (100, 1, 256, 256)
        # the image data part
        raw_image = (
            np.arange(np.prod(raw_data_shape)).reshape(raw_data_shape)
        ).astype(np.uint16)
        # keeping in mind 12-bit ADC
        raw_image = np.bitwise_and(raw_image, 0x0FFF)
        cls.raw_image_32 = raw_image.astype(np.float32)
        # the gain values (TODO: test with some being illegal)
        gain_data = (np.random.random(raw_data_shape) * 3).astype(np.uint16)
        assert np.array_equal(gain_data, np.bitwise_and(gain_data, 0x0003))
        cls.gain_data_32 = gain_data.astype(np.float32)
        cls.raw_data = np.bitwise_or(raw_image, np.left_shift(gain_data, 12))
        # just checking that we constructed something reasonable
        assert np.array_equal(
            gain_data, np.bitwise_and(np.right_shift(cls.raw_data, 12), 0x0003)
        )
        assert np.array_equal(raw_image, np.bitwise_and(cls.raw_data, 0x0FFF))

        # generate some defects
        cls.wrong_gain_value_indices = (
            np.random.random(raw_data_shape) < 0.005
        ).astype(np.bool_)
        raw_data_with_some_wrong_gain = cls.raw_data.copy()
        raw_data_with_some_wrong_gain[cls.wrong_gain_value_indices] = np.bitwise_or(
            raw_data_with_some_wrong_gain[cls.wrong_gain_value_indices], 0x3000
        )
        cls.raw_data_with_some_wrong_gain = raw_data_with_some_wrong_gain

        # generate some constants
        dark_constant_shape = (256, 256, 512, 3)
        # okay, as slow and fast scan have same size, can't tell the difference here...
        funky_constant_shape = (256, 256, 512, 3)
        cls.offset_constant = np.random.random(dark_constant_shape).astype(np.float32)
        cls.gain_amp_constant = np.random.random(funky_constant_shape).astype(
            np.float32
        )
        cls.bp_dark_constant = (np.random.random(dark_constant_shape) < 0.01).astype(
            np.uint32
        )

        cell_table = np.linspace(0, 512, 100).astype(np.uint16)
        np.random.shuffle(cell_table)
        cls.cell_table = cell_table

    def test_only_cast(self, kernel_runner):
        _, processed, previews = kernel_runner.correct(self.raw_data, self.cell_table)
        assert_allclose(processed, self.raw_image_32[:, 0])
        # TODO: make raw preview show image data (remove gain bits)
        assert_allclose(previews[0], self.raw_data[:, 0])  # raw: raw
        assert_allclose(previews[1], self.raw_image_32[:, 0])  # processed: raw
        assert_allclose(previews[2], self.gain_data_32[:, 0])  # gain: unchanged

    def test_only_cast_with_some_wrong_gain(self, kernel_runner):
        (
            _,
            processed,
            _,
        ) = kernel_runner.correct(self.raw_data_with_some_wrong_gain, self.cell_table)
        assert np.all(np.isnan(processed[self.wrong_gain_value_indices[:, 0]]))

    def test_correct(self, cpu_runner, gpu_runner):
        # naive numpy version way too slow, just compare the two runners
        # the CPU one uses kernel almost identical to pycalibration
        cpu_runner.load_constant(Constants.Offset, self.offset_constant)
        cpu_runner.load_constant(Constants.GainAmpMap, self.gain_amp_constant)
        cpu_runner.load_constant(Constants.BadPixelsDark, self.bp_dark_constant)
        gpu_runner.load_constant(Constants.Offset, self.offset_constant)
        gpu_runner.load_constant(Constants.GainAmpMap, self.gain_amp_constant)
        gpu_runner.load_constant(Constants.BadPixelsDark, self.bp_dark_constant)
        _, processed_cpu, _ = cpu_runner.correct(
            self.raw_data_with_some_wrong_gain, self.cell_table
        )
        _, processed_gpu, _ = gpu_runner.correct(
            self.raw_data_with_some_wrong_gain, self.cell_table
        )
        assert np.allclose(processed_cpu, processed_gpu.get(), equal_nan=True)
