import numpy as np
import pytest
from karabo.bound import Hash
from calng.preview_utils import PreviewFriend, PreviewSpec

from common_setup import DummyCorrectionDevice


cell_table = np.array([0, 2, 1], dtype=np.uint16)
pulse_table = np.array([1, 2, 0], dtype=np.uint16)
raw_data = np.array(
    [
        [1.0, 2.0, 3.0],
        [4.0, np.nan, 5.0],
        [6.0, 7.0, 8.0],
    ],
    dtype=np.float32,
)
raw_without_nan = raw_data.copy()
raw_without_nan[~np.isfinite(raw_data)] = 0


@pytest.fixture
def device():
    output_specs = [
        PreviewSpec(
            "dummyPreview",
            dimensions=1,
            wrap_in_imagedata=False,
            frame_reduction=True,
        )
    ]
    device = DummyCorrectionDevice(
        {"friend_class": PreviewFriend, "outputs": output_specs}
    )

    def writeChannel(name, hsh, **kwargs):
        if not hasattr(device, "_written"):
            device._written = []
        device._written.append(("name", hsh))

    device.writeChannel = writeChannel
    device.preview_friend = PreviewFriend(device, output_specs)

    return device


def test_preview_frame(device):
    device.preview_friend.reconfigure(
        Hash(
            "dummyPreview.index",
            1,
            "dummyPreview.selectionMode",
            "frame",
        )
    )
    device.preview_friend.write_outputs(
        raw_data, cell_table=cell_table, pulse_table=pulse_table
    )
    assert hasattr(device, "_written")
    assert len(device._written) == 1
    preview = device._written[0][1]
    assert np.allclose(preview["image.data"], raw_without_nan[1])


def test_preview_cell(device):
    device.preview_friend.reconfigure(
        Hash(
            "dummyPreview.index",
            1,
            "dummyPreview.selectionMode",
            "cell",
        )
    )
    device.preview_friend.write_outputs(
        raw_data, cell_table=cell_table, pulse_table=pulse_table
    )
    preview = device._written[0][1]
    assert np.allclose(preview["image.data"], raw_without_nan[2])


def test_preview_pulse(device):
    device.preview_friend.reconfigure(
        Hash(
            "dummyPreview.index",
            2,
            "dummyPreview.selectionMode",
            "pulse",
        )
    )
    device.preview_friend.write_outputs(
        raw_data, cell_table=cell_table, pulse_table=pulse_table
    )
    preview = device._written[0][1]
    assert np.allclose(preview["image.data"], raw_without_nan[1])


def test_preview_max(device):
    device.preview_friend.reconfigure(
        Hash(
            "dummyPreview.index",
            -1,
        )
    )
    device.preview_friend.write_outputs(
        raw_data, cell_table=cell_table, pulse_table=pulse_table
    )
    preview = device._written[0][1]
    assert np.allclose(preview["image.data"], np.nanmax(raw_data, axis=0))


def test_preview_mean(device):
    device.preview_friend.reconfigure(
        Hash(
            "dummyPreview.index",
            -2,
        )
    )
    device.preview_friend.write_outputs(
        raw_data, cell_table=cell_table, pulse_table=pulse_table
    )
    preview = device._written[0][1]
    assert np.allclose(
        preview["image.data"], np.nanmean(raw_data, axis=0, dtype=np.float32)
    )


def test_preview_sum(device):
    device.preview_friend.reconfigure(
        Hash(
            "dummyPreview.index",
            -3,
        )
    )
    device.preview_friend.write_outputs(
        raw_data, cell_table=cell_table, pulse_table=pulse_table
    )
    preview = device._written[0][1]
    assert np.allclose(
        preview["image.data"], np.nansum(raw_data, axis=0, dtype=np.float32)
    )


def test_preview_std(device):
    device.preview_friend.reconfigure(
        Hash(
            "dummyPreview.index",
            -4,
        )
    )
    device.preview_friend.write_outputs(
        raw_data, cell_table=cell_table, pulse_table=pulse_table
    )
    preview = device._written[0][1]
    assert np.allclose(
        preview["image.data"], np.nanstd(raw_data, axis=0, dtype=np.float32)
    )


# TODO: also test alternative preview index selection modes
